"""Unit tests for :mod:`~tmlt.core.domains.collections`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2022

import re
from collections.abc import Mapping
from contextlib import nullcontext as does_not_raise
from itertools import combinations_with_replacement
from test.unit.domains.abstract import DomainTests
from typing import Any, Callable, ContextManager, Dict, Optional, Type

import numpy as np
import pytest
from pyspark.sql.types import StringType

from tmlt.core.domains.base import Domain, OutOfDomainError
from tmlt.core.domains.collections import DictDomain, ListDomain
from tmlt.core.domains.numpy_domains import NumpyFloatDomain, NumpyIntegerDomain
from tmlt.core.utils.misc import get_fullname


class TestListDomain(DomainTests):
    """Tests for :class:`~tmlt.core.domains.collections.ListDomain`."""

    @pytest.fixture
    def domain_type(self) -> Type[Domain]:  # pylint: disable=no-self-use
        """Returns the type of the domain to be tested."""
        return ListDomain

    @pytest.fixture(scope="class")
    def domain(self) -> ListDomain:  # pylint: disable=no-self-use
        """Get a base ListDomain."""
        return ListDomain(NumpyIntegerDomain())

    @pytest.mark.parametrize(
        "domain_args, expectation, exception_properties",
        [
            (
                {"element_domain": invalid_type},
                pytest.raises(
                    TypeError,
                    match=f"type of element_domain must be {get_fullname(Domain)}; "
                    f"got {get_fullname(invalid_type)} instead",
                ),
                None,
            )
            for invalid_type in [None, "not a domain", StringType, np.int64(1)]
        ]
        + [
            (
                {"element_domain": NumpyIntegerDomain(), "length": -1},
                pytest.raises(ValueError, match="length must be non-negative"),
                None,
            ),
            (
                {"element_domain": NumpyFloatDomain(), "length": 1.5},
                pytest.raises(
                    TypeError,
                    match=re.escape(
                        f"type of length must be one of ({get_fullname(int)}, "
                        f"{get_fullname(None)}); got {get_fullname(float)} instead"
                    ),
                ),
                None,
            ),
            (
                {"element_domain": NumpyIntegerDomain(), "length": 1},
                does_not_raise(),
                None,
            ),
            (
                {"element_domain": NumpyIntegerDomain(), "length": np.int64(5)},
                pytest.raises(
                    TypeError,
                    match=re.escape(
                        f"type of length must be one of ({get_fullname(int)}, "
                        f"{get_fullname(None)}); got {get_fullname(np.int64)} instead"
                    ),
                ),
                None,
            ),
            ({"element_domain": NumpyIntegerDomain()}, does_not_raise(), None),
            (
                {"element_domain": ListDomain(NumpyIntegerDomain())},
                does_not_raise(),
                None,
            ),
        ],
    )
    def test_construct_component(
        self,
        domain_type: Type[Domain],
        domain_args: Dict[str, Any],
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Initialization behaves correctly.

        The domain is constructed correctly and raises exceptions when initialized with
        invalid inputs.

        Args:
            domain_type: The type of domain to be constructed.
            domain_args: The arguments to the domain.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        super().test_construct_component(
            domain_type, domain_args, expectation, exception_properties
        )

    @pytest.mark.parametrize(
        "domain, other_domain, expected",
        [
            (ListDomain(NumpyIntegerDomain()), NumpyFloatDomain(), False),
            (ListDomain(NumpyIntegerDomain()), ListDomain(NumpyFloatDomain()), False),
            (ListDomain(NumpyIntegerDomain()), ListDomain(NumpyIntegerDomain()), True),
        ]
        + [
            # Lengths should be equal as well
            (
                ListDomain(NumpyIntegerDomain(), length=len1),
                ListDomain(NumpyIntegerDomain(), length=len2),
                len1 == len2,
            )
            for len1, len2 in combinations_with_replacement([1, 2, 3], 2)
        ],
    )
    def test_eq(self, domain: Domain, other_domain: Domain, expected: bool):
        """__eq__ works correctly.

        Args:
            domain: The domain to test.
            other_domain: The domain to compare to.
            expected: The expected result of the comparison.
        """
        super().test_eq(domain, other_domain, expected)

    @pytest.mark.skip("No arguments to mutate")
    @pytest.mark.parametrize("domain_args, key, mutator", [])
    def test_mutable_inputs(
        self,
        domain_type: Type[Domain],
        domain_args: Dict[str, Any],
        key: str,
        mutator: Callable[[Any], Any],
    ):
        """The mutable inputs to the domain are copied.

        Args:
            domain_type: The type of domain to be constructed.
            domain_args: The arguments to the domain.
            key: The parameter name to be changed.
            mutator: A lambda function that mutates the parameter.
        """
        super().test_mutable_inputs(domain_type, domain_args, key, mutator)

    @pytest.mark.parametrize(
        "domain, expected_properties",
        [
            (
                ListDomain(NumpyIntegerDomain()),
                {
                    "element_domain": NumpyIntegerDomain(),
                    "carrier_type": list,
                    "length": None,
                },
            ),
            (
                ListDomain(NumpyFloatDomain()),
                {
                    "element_domain": NumpyFloatDomain(),
                    "carrier_type": list,
                    "length": None,
                },
            ),
            (
                ListDomain(NumpyIntegerDomain(), length=5),
                {
                    "element_domain": NumpyIntegerDomain(),
                    "carrier_type": list,
                    "length": 5,
                },
            ),
        ],
    )
    def test_properties(self, domain: Domain, expected_properties: Dict[str, Any]):
        """All properties have the expected values.

        Args:
            domain: The constructed domain to be tested.
            expected_properties: A dictionary containing all the property:value pairs
                domain is expected to have.
        """
        super().test_properties(domain, expected_properties)

    @pytest.mark.parametrize(
        "domain", [ListDomain(NumpyIntegerDomain()), ListDomain(NumpyFloatDomain())]
    )
    def test_property_immutability(self, domain: Domain):
        """The properties return copies for mutable values.

        Args:
            domain: The domain to be tested.
        """
        super().test_property_immutability(domain)

    @pytest.mark.parametrize(
        "domain, candidate, expectation, exception_properties",
        [
            (ListDomain(NumpyIntegerDomain()), [np.int64(1)], does_not_raise(), None),
            (
                ListDomain(NumpyIntegerDomain()),
                [np.float64(1.0)],
                pytest.raises(
                    OutOfDomainError,
                    match="Found invalid value in list: Value must be "
                    f"{get_fullname(np.int64)}, "
                    f"instead it is {get_fullname(np.float64)}",
                ),
                {
                    "domain": ListDomain(NumpyIntegerDomain()),
                    "value": [np.float64(1.0)],
                },
            ),
            (
                ListDomain(NumpyIntegerDomain()),
                "not a list",
                pytest.raises(
                    OutOfDomainError,
                    match=f"Value must be {get_fullname(list)}, instead it is "
                    f"{get_fullname(str)}.",
                ),
                {"domain": ListDomain(NumpyIntegerDomain()), "value": "not a list"},
            ),
            (
                ListDomain(NumpyIntegerDomain(), length=3),
                [np.int64(i) for i in range(10)],
                pytest.raises(
                    OutOfDomainError,
                    match=f"Expected list of length {3}, found list of length {10}",
                ),
                {
                    "domain": ListDomain(NumpyIntegerDomain(), length=3),
                    "value": [np.int64(i) for i in range(10)],
                },
            ),
            (
                ListDomain(NumpyIntegerDomain(), length=10),
                [np.int64(i) for i in range(10)],
                does_not_raise(),
                None,
            ),
        ],
    )
    def test_validate(
        self,
        domain: Domain,
        candidate: Any,
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Validate works correctly.

        Args:
            domain: The domain to test.
            candidate: The value to validate using domain.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        super().test_validate(domain, candidate, expectation, exception_properties)


class TestDictDomain(DomainTests):
    """Tests for :class:`~tmlt.core.domains.collections.DictDomain`."""

    @pytest.fixture
    def domain_type(self) -> Type[Domain]:  # pylint: disable=no-self-use
        """Returns the type of the domain to be tested."""
        return DictDomain

    @pytest.fixture(scope="class")
    def domain(self) -> DictDomain:  # pylint: disable=no-self-use
        """Get a base DictDomain."""
        return DictDomain({"A": NumpyIntegerDomain(), "B": NumpyFloatDomain()})

    @pytest.mark.parametrize(
        "domain_args, expectation, exception_properties",
        [
            (
                {"key_to_domain": []},
                pytest.raises(
                    TypeError,
                    match='type of argument "key_to_domain" must be '
                    f"{get_fullname(Mapping)}; got {get_fullname(list)} instead",
                ),
                None,
            ),
            (
                {"key_to_domain": (1, 2, 3)},
                pytest.raises(
                    TypeError,
                    match='type of argument "key_to_domain" must be '
                    f"{get_fullname(Mapping)}; got {get_fullname(tuple)} instead",
                ),
                None,
            ),
            (
                {"key_to_domain": "not a domain"},
                pytest.raises(
                    TypeError,
                    match='type of argument "key_to_domain" must be '
                    f"{get_fullname(Mapping)}; got {get_fullname(str)} instead",
                ),
                None,
            ),
            (
                {"key_to_domain": {"A": np.int64(1)}},
                pytest.raises(
                    TypeError,
                    match=f"Expected domain for key 'A' to be a {get_fullname(Domain)};"
                    f" got {get_fullname(np.int64)} instead",
                ),
                None,
            ),
            (
                {"key_to_domain": {"A": 1}},
                pytest.raises(
                    TypeError,
                    match=f"Expected domain for key 'A' to be a {get_fullname(Domain)};"
                    f" got {get_fullname(int)} instead",
                ),
                None,
            ),
            ({"key_to_domain": {1: NumpyIntegerDomain()}}, does_not_raise(), None),
            (
                {"key_to_domain": {"A": ListDomain(NumpyIntegerDomain())}},
                does_not_raise(),
                None,
            ),
            (
                {
                    "key_to_domain": {
                        "A": DictDomain({"B": DictDomain({"C": NumpyIntegerDomain()})})
                    }
                },
                does_not_raise(),
                None,
            ),
        ]
        + [
            # Testing complex keys
            ({"key_to_domain": {key: NumpyIntegerDomain()}}, does_not_raise(), None)
            for key in [("A",), "3", 1.0, np.int64(-3)]
        ],
    )
    def test_construct_component(
        self,
        domain_type: Type[Domain],
        domain_args: Dict[str, Any],
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Initialization behaves correctly.

        The domain is constructed correctly and raises exceptions when initialized with
        invalid inputs.

        Args:
            domain_type: The type of domain to be constructed.
            domain_args: The arguments to the domain.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        super().test_construct_component(
            domain_type, domain_args, expectation, exception_properties
        )

    @pytest.mark.parametrize(
        "other_domain, expected",
        [
            (DictDomain({"A": NumpyIntegerDomain(), "B": NumpyFloatDomain()}), True),
            (DictDomain({"B": NumpyFloatDomain(), "A": NumpyIntegerDomain()}), True),
            (
                DictDomain(
                    {"A": NumpyIntegerDomain(), "B": NumpyFloatDomain(allow_nan=True)}
                ),
                False,
            ),
            (DictDomain({"A": NumpyIntegerDomain(), "C": NumpyFloatDomain()}), False),
            (
                DictDomain(
                    {
                        "A": NumpyIntegerDomain(),
                        "B": NumpyFloatDomain(),
                        "C": NumpyFloatDomain(),
                    }
                ),
                False,
            ),
            (DictDomain({"A": NumpyIntegerDomain()}), False),
            (NumpyIntegerDomain(), False),
        ],
    )
    def test_eq(self, domain: Domain, other_domain: Domain, expected: bool):
        """__eq__ works correctly.

        Args:
            domain: The domain to test.
            other_domain: The domain to compare to.
            expected: The expected result of the comparison.
        """
        super().test_eq(domain, other_domain, expected)

    @pytest.mark.parametrize(
        "domain_args, key, mutator",
        [
            (
                {"key_to_domain": {"A": NumpyIntegerDomain()}},
                "key_to_domain",
                lambda x: x.update({"A": NumpyFloatDomain()}),
            )
        ],
    )
    def test_mutable_inputs(
        self,
        domain_type: Type[Domain],
        domain_args: Dict[str, Any],
        key: str,
        mutator: Callable[[Any], Any],
    ):
        """The mutable inputs to the domain are copied.

        Args:
            domain_type: The type of domain to be constructed.
            domain_args: The arguments to the domain.
            key: The parameter name to be changed.
            mutator: A lambda function that mutates the parameter.
        """
        super().test_mutable_inputs(domain_type, domain_args, key, mutator)

    @pytest.mark.parametrize(
        "domain, expected_properties",
        [
            (DictDomain({}), {"key_to_domain": {}, "length": 0, "carrier_type": dict}),
            (
                DictDomain({"A": NumpyIntegerDomain()}),
                {
                    "key_to_domain": {"A": NumpyIntegerDomain()},
                    "length": 1,
                    "carrier_type": dict,
                },
            ),
            (
                DictDomain({"A": NumpyIntegerDomain(), "B": NumpyFloatDomain()}),
                {
                    "key_to_domain": {
                        "A": NumpyIntegerDomain(),
                        "B": NumpyFloatDomain(),
                    },
                    "length": 2,
                    "carrier_type": dict,
                },
            ),
        ],
    )
    def test_properties(self, domain: Domain, expected_properties: Dict[str, Any]):
        """All properties have the expected values.

        Args:
            domain: The constructed domain to be tested.
            expected_properties: A dictionary containing all the property:value pairs
                domain is expected to have.
        """
        super().test_properties(domain, expected_properties)

    @pytest.mark.parametrize(
        "domain", [(DictDomain({"A": NumpyIntegerDomain(), "B": NumpyFloatDomain()}))]
    )
    def test_property_immutability(self, domain: Domain):
        """The properties return copies for mutable values.

        Args:
            domain: The domain to be tested.
        """
        super().test_property_immutability(domain)

    @pytest.mark.parametrize(
        "domain, candidate, expectation, exception_properties",
        [
            (
                DictDomain({"A": NumpyIntegerDomain()}),
                NumpyIntegerDomain(),
                pytest.raises(
                    OutOfDomainError,
                    match=f"Value must be {get_fullname(dict)}, instead it is "
                    f"{get_fullname(NumpyIntegerDomain)}.",
                ),
                {"domain": DictDomain({"A": NumpyIntegerDomain()})},
            ),
            (
                DictDomain({"A": NumpyIntegerDomain()}),
                {"A": np.int64(1)},
                does_not_raise(),
                None,
            ),
            (
                DictDomain({"A": NumpyIntegerDomain()}),
                {"A": np.int64(1), "B": np.int64(2)},
                pytest.raises(
                    OutOfDomainError,
                    match=re.escape(
                        "Keys are not as expected, value must match domain.\n"
                        "Value keys: ['A', 'B']\nDomain keys: ['A']"
                    ),
                ),
                {
                    "domain": DictDomain({"A": NumpyIntegerDomain()}),
                    "value": ["A", "B"],
                },
            ),
            (
                DictDomain({"A": NumpyIntegerDomain(), "B": NumpyFloatDomain()}),
                {"A": np.int64(1)},
                pytest.raises(
                    OutOfDomainError,
                    match=re.escape(
                        "Keys are not as expected, value must match domain.\n"
                        "Value keys: ['A']\nDomain keys: ['A', 'B']"
                    ),
                ),
                {
                    "domain": DictDomain(
                        {"A": NumpyIntegerDomain(), "B": NumpyFloatDomain()}
                    ),
                    "value": ["A"],
                },
            ),
            (
                DictDomain({"A": NumpyIntegerDomain(), "B": NumpyFloatDomain()}),
                {"A": np.float64(1.0), "B": np.float64(2.0)},
                pytest.raises(
                    OutOfDomainError,
                    match=f"Found invalid value at 'A': Value must be "
                    f"{get_fullname(np.int64)}, instead it is "
                    f"{get_fullname(np.float64)}.",
                ),
                {
                    "domain": DictDomain(
                        {"A": NumpyIntegerDomain(), "B": NumpyFloatDomain()}
                    ),
                    "value": {"A": np.float64(1.0), "B": np.float64(2.0)},
                },
            ),
            (DictDomain({}), {}, does_not_raise(), None),
            (
                DictDomain({}),
                {"A": np.int64(1), "B": np.float64(2.0)},
                pytest.raises(
                    OutOfDomainError,
                    match=re.escape(
                        "Keys are not as expected, value must match domain.\n"
                        "Value keys: ['A', 'B']\nDomain keys: []"
                    ),
                ),
                {"domain": DictDomain({}), "value": ["A", "B"]},
            ),
        ],
    )
    def test_validate(
        self,
        domain: Domain,
        candidate: Any,
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Validate works correctly.

        Args:
            domain: The domain to test.
            candidate: The value to validate using domain.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        super().test_validate(domain, candidate, expectation, exception_properties)



"""Unit tests for :mod:`~tmlt.core.measurements.noise_mechanisms`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

import re
from contextlib import nullcontext as does_not_raise
from test.unit.measurements.abstract import MeasurementTests
from typing import Any, Callable, ContextManager, Dict, Type

import numpy as np
import pytest
import sympy as sp
from pyspark.sql.types import DoubleType, LongType

from tmlt.core.domains.numpy_domains import (
    NumpyFloatDomain,
    NumpyIntegerDomain,
    NumpyStringDomain,
)
from tmlt.core.measurements.base import Measurement
from tmlt.core.measurements.noise_mechanisms import (
    AddDiscreteGaussianNoise,
    AddGaussianNoise,
    AddGeometricNoise,
    AddLaplaceNoise,
)
from tmlt.core.measures import PureDP, RhoZCDP
from tmlt.core.metrics import AbsoluteDifference
from tmlt.core.utils.exact_number import ExactNumber
from tmlt.core.utils.misc import get_fullname

# pylint: disable=no-self-use


class TestAddLaplaceNoise(MeasurementTests):
    """Tests for :class:`~tmlt.core.measurements.noise_mechanisms.AddLaplaceNoise`."""

    @pytest.fixture
    def measurement_type(self) -> Type[Measurement]:  # pylint: disable=no-self-use
        """Returns the type of the measurement to be tested."""
        return AddLaplaceNoise

    @pytest.mark.parametrize(
        "measurement_args, expectation",
        [
            (
                {"input_domain": NumpyIntegerDomain(), "scale": sp.Rational("0.3")},
                does_not_raise(),
            ),
            (
                {
                    "input_domain": NumpyIntegerDomain(size=32),
                    "scale": sp.Rational("0.3"),
                },
                does_not_raise(),
            ),
            (
                {"input_domain": NumpyStringDomain(), "scale": 1},
                pytest.raises(
                    TypeError,
                    match=re.escape(
                        'type of argument "input_domain" must be one of '
                        "(NumpyIntegerDomain, NumpyFloatDomain); got "
                        f"{get_fullname(NumpyStringDomain)} instead"
                    ),
                ),
            ),
            (
                {"input_domain": NumpyIntegerDomain(), "scale": "x"},
                pytest.raises(
                    ValueError, match="Invalid scale: x contains free symbols"
                ),
            ),
            (
                {"input_domain": NumpyIntegerDomain(), "scale": -1},
                pytest.raises(
                    ValueError,
                    match="Invalid scale: -1 is not greater than or equal to 0",
                ),
            ),
            ({"input_domain": NumpyFloatDomain(), "scale": "52/3"}, does_not_raise()),
            (
                {"input_domain": NumpyFloatDomain(size=32), "scale": "52/3"},
                does_not_raise(),
            ),
            (
                {"input_domain": NumpyFloatDomain(allow_nan=True), "scale": "3 + 2"},
                pytest.raises(
                    ValueError, match="Input domain must not contain infs or nans"
                ),
            ),
            (
                {
                    "input_domain": NumpyFloatDomain(allow_inf=True),
                    "scale": ExactNumber(3),
                },
                pytest.raises(
                    ValueError, match="Input domain must not contain infs or nans"
                ),
            ),
        ],
    )
    def test_construct_component(
        self,
        measurement_type: Type[Measurement],
        measurement_args: Dict[str, Any],
        expectation: ContextManager[None],
    ):
        """Initialization behaves correctly.

        The measurement is constructed correctly and raises exceptions
        when initialized with invalid inputs.

        Args:
            measurement_type: The type of the measurement to be constructed.
            measurement_args: The arguments to the measurement.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        super().test_construct_component(
            measurement_type, measurement_args, expectation, exception_properties=None
        )

    @pytest.mark.parametrize(
        "measurement, input_data",
        [
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=float("inf")),
                np.int64(1),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyFloatDomain(), scale=float("inf")),
                np.float64(-1),
            ),
        ],
    )
    def test_infinite_protection(self, measurement: Measurement, input_data: Any):
        """Measurement works with infinite protection.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
        """
        output = measurement(input_data)
        assert output in [float("inf"), -float("inf")]

    @pytest.mark.skip("No arguments to mutate")
    def test_mutable_inputs(
        self,
        measurement_type: Type[Measurement],
        measurement_args: Dict[str, Any],
        key: str,
        mutator: Callable,
    ):
        """Mutable inputs to the measurement are copied.

        Args:
            measurement_type: The type of the measurement to be constructed.
            measurement_args: The arguments to the measurement.
            key: The parameter name to be changed.
            mutator: A lambda function that mutates the parameter.
        """
        super().test_mutable_inputs(measurement_type, measurement_args, key, mutator)

    @pytest.mark.parametrize(
        "measurement, input_data, expected_output",
        [
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=0),
                np.int64(10),
                10.0,
            ),
            (
                AddLaplaceNoise(input_domain=NumpyFloatDomain(size=32), scale=0),
                np.float32(-0.1234567),
                float(np.float32(-0.1234567)),
            ),
        ],
    )
    def test_no_protection(
        self, measurement: Measurement, input_data: Any, expected_output: Any
    ):
        """Measurement works with no protection.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
            expected_output: Expected unprotected output.
        """
        super().test_no_protection(measurement, input_data, expected_output)

    @pytest.mark.parametrize(
        "measurement, input_data, expected_output_type",
        [
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=1),
                np.int64(10),
                float,
            ),
            (
                AddLaplaceNoise(input_domain=NumpyFloatDomain(size=32), scale=1),
                np.float32(-0.1234567),
                float,
            ),
        ],
    )
    def test_output(
        self, measurement: Measurement, input_data: Any, expected_output_type: type
    ):
        """Measurement produces an output that has the expected type.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
            expected_output_type: The expected type for the output.
        """
        super().test_output(measurement, input_data, expected_output_type)

    @pytest.mark.parametrize(
        "measurement, d_in, expected_d_out, expectation",
        [
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=1),
                1,
                1,
                does_not_raise(),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=2),
                1,
                "0.5",
                does_not_raise(),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyFloatDomain(), scale="0.5"),
                1,
                2,
                does_not_raise(),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyFloatDomain(), scale="0.5"),
                3,
                6,
                does_not_raise(),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyFloatDomain(), scale="0.5"),
                -1,
                None,
                pytest.raises(
                    ValueError,
                    match="Invalid value for metric AbsoluteDifference: -1 is not "
                    "greater than or equal to 0",
                ),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=0),
                1,
                float("inf"),
                does_not_raise(),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=0),
                0,
                0,
                does_not_raise(),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=float("inf")),
                1,
                0,
                does_not_raise(),
            ),
        ],
    )
    def test_privacy_function(
        self,
        measurement: Measurement,
        d_in: Any,
        expected_d_out: Any,
        expectation: ContextManager[None],
    ):
        """Testing the measurement's privacy function.

        Args:
            measurement: The measurement to be tested.
            d_in: Distance between inputs.
            expected_d_out: Expected distance between outputs.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        super().test_privacy_function(
            measurement, d_in, expected_d_out, expectation, exception_properties=None
        )

    @pytest.mark.parametrize(
        "measurement, d_in, d_out, expected, expectation",
        [
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=1),
                1,
                1,
                True,
                does_not_raise(),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=1),
                1,
                2,
                True,
                does_not_raise(),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=1),
                1,
                "0.5",
                False,
                does_not_raise(),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyFloatDomain(), scale="0.5"),
                3,
                6,
                True,
                does_not_raise(),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyFloatDomain(), scale="0.5"),
                3,
                5,
                False,
                does_not_raise(),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyFloatDomain(), scale="0.5"),
                -1,
                1,
                True,
                pytest.raises(
                    ValueError,
                    match="Invalid value for metric AbsoluteDifference: -1 is not "
                    "greater than or equal to 0",
                ),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyFloatDomain(), scale="0.5"),
                tuple(),
                1,
                True,
                pytest.raises(
                    TypeError,
                    match=re.escape(
                        'type of argument "d_in" must be one of (ExactNumber, float, '
                        "int, str, Fraction, Expr); got tuple instead"
                    ),
                ),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyFloatDomain(), scale=1),
                1,
                -1,
                True,
                pytest.raises(
                    ValueError,
                    match=re.escape(
                        "Invalid PureDP measure value (epsilon): -1 is not greater "
                        "than or equal to 0"
                    ),
                ),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyFloatDomain(), scale=1),
                1,
                {},
                True,
                pytest.raises(
                    TypeError,
                    match=re.escape(
                        'type of argument "value" must be one of (ExactNumber, float, '
                        "int, str, Fraction, Expr); got dict instead"
                    ),
                ),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=0),
                1,
                float("inf"),
                True,
                does_not_raise(),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=0),
                0,
                0,
                True,
                does_not_raise(),
            ),
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=float("inf")),
                1,
                0,
                True,
                does_not_raise(),
            ),
        ],
    )
    def test_privacy_relation(
        self,
        measurement: Measurement,
        d_in: Any,
        d_out: Any,
        expected: bool,
        expectation: ContextManager[None],
    ):
        """Testing the measurement's privacy function.

        Args:
            measurement: The measurement to be tested.
            d_in: Distance between inputs.
            d_out: Distance between outputs.
            expected: Whether the d_in, d_out pair is close.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        super().test_privacy_relation(
            measurement, d_in, d_out, expected, expectation, exception_properties=None
        )

    @pytest.mark.parametrize(
        "measurement, expected_properties",
        [
            (
                AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=1),
                {
                    "input_domain": NumpyIntegerDomain(),
                    "scale": 1,
                    "input_metric": AbsoluteDifference(),
                    "output_type": DoubleType(),
                    "output_measure": PureDP(),
                    "is_interactive": False,
                },
            ),
            (
                AddLaplaceNoise(input_domain=NumpyFloatDomain(size=32), scale=2),
                {
                    "input_domain": NumpyFloatDomain(size=32),
                    "scale": 2,
                    "input_metric": AbsoluteDifference(),
                    "output_type": DoubleType(),
                    "output_measure": PureDP(),
                    "is_interactive": False,
                },
            ),
        ],
    )
    def test_properties(
        self, measurement: Measurement, expected_properties: Dict[str, Any]
    ):
        """All properties have the expected values.

        Args:
            measurement: The constructed measurement to be tested.
            expected_properties: A dictionary containing all the
                property:value pairs measurement is expected to have.
        """
        super().test_properties(measurement, expected_properties)

    @pytest.mark.parametrize(
        "measurement", [AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=1)]
    )
    def test_property_immutability(self, measurement: Measurement):
        """The properties return copies for mutable values.

        Args:
            measurement: The measurement to be tested.
        """
        super().test_property_immutability(measurement)

    @pytest.mark.parametrize(
        "scale, p, expected, expectation",
        [
            (1, 0.5, 0, does_not_raise()),
            (2, 0.5, 0, does_not_raise()),
            (3.1415, 0.5, 0, does_not_raise()),
            (1, 0.75, 0.693147, does_not_raise()),
            (1, 0.25, -0.693147, does_not_raise()),
            (3.5, 0.86, 4.45538, does_not_raise()),
            (3.5, 0.14, -4.45538, does_not_raise()),
            (0, 0.6, 0, does_not_raise()),  # 0 scale is 0
            (  # inf scale is +inf 50% of the time
                float("inf"),
                0.6,
                float("inf"),
                does_not_raise(),
            ),
            (  # inf scale is -inf 50% of the time
                float("inf"),
                0.4,
                float("-inf"),
                does_not_raise(),
            ),
            (1, 0, float("-inf"), does_not_raise()),  # p=0 is -inf
            (1, 1, float("inf"), does_not_raise()),  # p=1 is +inf
            (-1, 0.6, float("nan"), does_not_raise()),  # negative scale is nan
            (float("nan"), 0.6, float("nan"), does_not_raise()),  # nan scale is nan
        ]
        + [  # p values outside [0, 1] raise ValueError
            (
                1,
                bad_p,
                float("nan"),
                pytest.raises(
                    ValueError,
                    match=re.escape(
                        "Probabilities input to the inverse CDF must be in [0, 1], "
                        f"but got {bad_p}"
                    ),
                ),
            )
            for bad_p in [-1, 1.1, float("inf"), float("nan")]
        ],
    )
    def test_inverse_cdf(
        self, scale: float, p: float, expected: float, expectation: ContextManager[None]
    ):
        """Testing the inverse_cdf function.

        Args:
            scale: The scale of the Laplace distribution.
            p: The probability.
            expected: The expected value.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        with expectation:
            actual = AddLaplaceNoise.inverse_cdf(scale=scale, probability=p)
            if np.isnan(expected):
                assert np.isnan(actual), f"Expected {expected}, got {actual}"
            else:
                assert actual == pytest.approx(
                    expected
                ), f"Expected {expected}, got {actual}"


class TestAddGeometricNoise(MeasurementTests):
    """Tests for :class:`~tmlt.core.measurements.noise_mechanisms.AddGeometricNoise`."""

    @pytest.fixture
    def measurement_type(self) -> Type[Measurement]:  # pylint: disable=no-self-use
        """Returns the type of the measurement to be tested."""
        return AddGeometricNoise

    @pytest.mark.parametrize(
        "measurement_args, expectation",
        [
            ({"alpha": sp.Rational("0.3")}, does_not_raise()),
            (
                {"alpha": "x"},
                pytest.raises(
                    ValueError, match="Invalid alpha: x contains free symbols"
                ),
            ),
            (
                {"alpha": -1},
                pytest.raises(
                    ValueError,
                    match="Invalid alpha: -1 is not greater than or equal to 0",
                ),
            ),
            (
                {"alpha": float("inf")},
                pytest.raises(
                    ValueError, match="Invalid alpha: inf is not strictly less than inf"
                ),
            ),
            ({"alpha": "0"}, does_not_raise()),
        ],
    )
    def test_construct_component(
        self,
        measurement_type: Type[Measurement],
        measurement_args: Dict[str, Any],
        expectation: ContextManager[None],
    ):
        """Initialization behaves correctly.

        The measurement is constructed correctly and raises exceptions
        when initialized with invalid inputs.

        Args:
            measurement_type: The type of the measurement to be constructed.
            measurement_args: The arguments to the measurement.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        super().test_construct_component(
            measurement_type, measurement_args, expectation, exception_properties=None
        )

    @pytest.mark.skip("Doesn't support infinite protection")
    def test_infinite_protection(
        self, measurement: Measurement, input_data: Any, expected_output: Any
    ):
        """Measurement works with infinite protection.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
            expected_output: The expected output of the measurement.
        """
        super().test_infinite_protection(measurement, input_data, expected_output)

    @pytest.mark.skip("No arguments to mutate")
    def test_mutable_inputs(
        self,
        measurement_type: Type[Measurement],
        measurement_args: Dict[str, Any],
        key: str,
        mutator: Callable,
    ):
        """Mutable inputs to the measurement are copied.

        Args:
            measurement_type: The type of the measurement to be constructed.
            measurement_args: The arguments to the measurement.
            key: The parameter name to be changed.
            mutator: A lambda function that mutates the parameter.
        """
        super().test_mutable_inputs(measurement_type, measurement_args, key, mutator)

    @pytest.mark.parametrize(
        "measurement, input_data, expected_output",
        [
            (AddGeometricNoise(alpha=0), np.int64(10), 10),
            (AddGeometricNoise(alpha=0), np.int64(-10), -10),
        ],
    )
    def test_no_protection(
        self, measurement: Measurement, input_data: Any, expected_output: Any
    ):
        """Measurement works with no protection.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
            expected_output: Expected unprotected output.
        """
        super().test_no_protection(measurement, input_data, expected_output)

    @pytest.mark.parametrize(
        "measurement, input_data, expected_output_type",
        [(AddGeometricNoise(alpha=3), np.int64(0), int)],
    )
    def test_output(
        self, measurement: Measurement, input_data: Any, expected_output_type: type
    ):
        """Measurement produces an output that has the expected type.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
            expected_output_type: The expected type for the output.
        """
        super().test_output(measurement, input_data, expected_output_type)

    @pytest.mark.parametrize(
        "measurement, d_in, expected_d_out, expectation",
        [
            (AddGeometricNoise(alpha=1), 1, 1, does_not_raise()),
            (AddGeometricNoise(alpha=2), 1, "0.5", does_not_raise()),
            (AddGeometricNoise(alpha="0.5"), 1, 2, does_not_raise()),
            (AddGeometricNoise(alpha="0.5"), 3, 6, does_not_raise()),
            (
                AddGeometricNoise(alpha="0.5"),
                -1,
                None,
                pytest.raises(
                    ValueError,
                    match="Invalid value for metric AbsoluteDifference: -1 is not "
                    "greater than or equal to 0",
                ),
            ),
            (AddGeometricNoise(alpha=0), 1, float("inf"), does_not_raise()),
            (AddGeometricNoise(alpha=0), 0, 0, does_not_raise()),
        ],
    )
    def test_privacy_function(
        self,
        measurement: Measurement,
        d_in: Any,
        expected_d_out: Any,
        expectation: ContextManager[None],
    ):
        """Testing the measurement's privacy function.

        Args:
            measurement: The measurement to be tested.
            d_in: Distance between inputs.
            expected_d_out: Expected distance between outputs.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        super().test_privacy_function(
            measurement, d_in, expected_d_out, expectation, exception_properties=None
        )

    @pytest.mark.parametrize(
        "measurement, d_in, d_out, expected, expectation",
        [
            (AddGeometricNoise(alpha=1), 1, 1, True, does_not_raise()),
            (AddGeometricNoise(alpha=1), 1, 2, True, does_not_raise()),
            (AddGeometricNoise(alpha=1), 1, "0.5", False, does_not_raise()),
            (AddGeometricNoise(alpha="0.5"), 3, 6, True, does_not_raise()),
            (AddGeometricNoise(alpha="0.5"), 3, 5, False, does_not_raise()),
            (
                AddGeometricNoise(alpha="0.5"),
                -1,
                1,
                True,
                pytest.raises(
                    ValueError,
                    match="Invalid value for metric AbsoluteDifference: -1 is not "
                    "greater than or equal to 0",
                ),
            ),
            (
                AddGeometricNoise(alpha="0.5"),
                tuple(),
                1,
                True,
                pytest.raises(
                    TypeError,
                    match=re.escape(
                        'type of argument "d_in" must be one of (ExactNumber, float, '
                        "int, str, Fraction, Expr); got tuple instead"
                    ),
                ),
            ),
            (
                AddGeometricNoise(alpha="0.5"),
                1,
                -1,
                True,
                pytest.raises(
                    ValueError,
                    match=re.escape(
                        "Invalid PureDP measure value (epsilon): -1 is not greater "
                        "than or equal to 0"
                    ),
                ),
            ),
            (
                AddGeometricNoise(alpha="0.5"),
                1,
                {},
                True,
                pytest.raises(
                    TypeError,
                    match=re.escape(
                        'type of argument "value" must be one of (ExactNumber, float, '
                        "int, str, Fraction, Expr); got dict instead"
                    ),
                ),
            ),
            (AddGeometricNoise(alpha=0), 1, float("inf"), True, does_not_raise()),
            (AddGeometricNoise(alpha=0), 0, 0, True, does_not_raise()),
        ],
    )
    def test_privacy_relation(
        self,
        measurement: Measurement,
        d_in: Any,
        d_out: Any,
        expected: bool,
        expectation: ContextManager[None],
    ):
        """Testing the measurement's privacy function.

        Args:
            measurement: The measurement to be tested.
            d_in: Distance between inputs.
            d_out: Distance between outputs.
            expected: Whether the d_in, d_out pair is close.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        super().test_privacy_relation(
            measurement, d_in, d_out, expected, expectation, exception_properties=None
        )

    @pytest.mark.parametrize(
        "measurement, expected_properties",
        [
            (
                AddGeometricNoise(alpha=1),
                {
                    "input_domain": NumpyIntegerDomain(),
                    "alpha": 1,
                    "input_metric": AbsoluteDifference(),
                    "output_type": LongType(),
                    "output_measure": PureDP(),
                    "is_interactive": False,
                },
            ),
            (
                AddGeometricNoise(alpha=2),
                {
                    "input_domain": NumpyIntegerDomain(),
                    "alpha": 2,
                    "input_metric": AbsoluteDifference(),
                    "output_type": LongType(),
                    "output_measure": PureDP(),
                    "is_interactive": False,
                },
            ),
        ],
    )
    def test_properties(
        self, measurement: Measurement, expected_properties: Dict[str, Any]
    ):
        """All properties have the expected values.

        Args:
            measurement: The constructed measurement to be tested.
            expected_properties: A dictionary containing all the
                property:value pairs measurement is expected to have.
        """
        super().test_properties(measurement, expected_properties)

    @pytest.mark.parametrize("measurement", [AddGeometricNoise(alpha=1)])
    def test_property_immutability(self, measurement: Measurement):
        """The properties return copies for mutable values.

        Args:
            measurement: The measurement to be tested.
        """
        super().test_property_immutability(measurement)

    @pytest.mark.parametrize(
        "alpha, p, expected, expectation",
        [
            (1, 0.5, 0, does_not_raise()),
            (2, 0.5, 0, does_not_raise()),
            (3.1415, 0.5, 0, does_not_raise()),
            (1, 0.75, 1, does_not_raise()),
            (1, 0.25, -1, does_not_raise()),
            (3.5, 0.86, 4, does_not_raise()),
            (3.5, 0.14, -4, does_not_raise()),
            (0, 0.6, 0, does_not_raise()),  # 0 scale is 0
            (  # inf scale is +inf 50% of the time
                float("inf"),
                0.6,
                float("inf"),
                does_not_raise(),
            ),
            (  # inf scale is -inf 50% of the time
                float("inf"),
                0.4,
                float("-inf"),
                does_not_raise(),
            ),
            (1, 0, float("-inf"), does_not_raise()),  # p=0 is -inf
            (1, 1, float("inf"), does_not_raise()),  # p=1 is +inf
            (-1, 0.6, float("nan"), does_not_raise()),  # negative scale is nan
            (float("nan"), 0.6, float("nan"), does_not_raise()),  # nan scale is nan
        ]
        + [  # p values outside [0, 1] raise ValueError
            (
                1,
                bad_p,
                float("nan"),
                pytest.raises(
                    ValueError,
                    match=re.escape(
                        "Probabilities input to the inverse CDF must be in [0, 1], "
                        f"but got {bad_p}"
                    ),
                ),
            )
            for bad_p in [-1, 1.1, float("inf"), float("nan")]
        ],
    )
    def test_inverse_cdf(
        self, alpha: float, p: float, expected: float, expectation: ContextManager[None]
    ):
        """Testing the inverse_cdf function.

        Args:
            alpha: The alpha of the Geometric distribution.
            p: The probability.
            expected: The expected value.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        with expectation:
            actual = AddGeometricNoise.inverse_cdf(alpha=alpha, probability=p)
            if np.isnan(expected):
                assert np.isnan(actual), f"Expected {expected}, got {actual}"
            else:
                assert actual == pytest.approx(
                    expected
                ), f"Expected {expected}, got {actual}"


class TestAddGaussianNoise(MeasurementTests):
    """Tests for :class:`~tmlt.core.measurements.noise_mechanisms.AddGaussianNoise`."""

    @pytest.fixture
    def measurement_type(self) -> Type[Measurement]:  # pylint: disable=no-self-use
        """Returns the type of the measurement to be tested."""
        return AddGaussianNoise

    @pytest.mark.parametrize(
        "measurement_args, expectation",
        [
            (
                {
                    "input_domain": NumpyIntegerDomain(),
                    "sigma_squared": sp.Rational("0.3"),
                },
                does_not_raise(),
            ),
            (
                {
                    "input_domain": NumpyIntegerDomain(size=32),
                    "sigma_squared": sp.Rational("0.3"),
                },
                does_not_raise(),
            ),
            (
                {"input_domain": NumpyStringDomain(), "sigma_squared": 1},
                pytest.raises(
                    TypeError,
                    match=re.escape(
                        'type of argument "input_domain" must be one of ('
                        "NumpyIntegerDomain, NumpyFloatDomain); got "
                        f"{get_fullname(NumpyStringDomain)} instead"
                    ),
                ),
            ),
            (
                {"input_domain": NumpyIntegerDomain(), "sigma_squared": "x"},
                pytest.raises(
                    ValueError, match="Invalid sigma_squared: x contains free symbols"
                ),
            ),
            (
                {"input_domain": NumpyIntegerDomain(), "sigma_squared": -1},
                pytest.raises(
                    ValueError,
                    match="Invalid sigma_squared: -1 is not greater than or equal to 0",
                ),
            ),
            (
                {"input_domain": NumpyFloatDomain(), "sigma_squared": "52/3"},
                does_not_raise(),
            ),
            (
                {"input_domain": NumpyFloatDomain(size=32), "sigma_squared": "52/3"},
                does_not_raise(),
            ),
            (
                {
                    "input_domain": NumpyFloatDomain(allow_nan=True),
                    "sigma_squared": "3 + 2",
                },
                pytest.raises(
                    ValueError, match="Input domain must not contain infs or nans"
                ),
            ),
            (
                {
                    "input_domain": NumpyFloatDomain(allow_inf=True),
                    "sigma_squared": ExactNumber(3),
                },
                pytest.raises(
                    ValueError, match="Input domain must not contain infs or nans"
                ),
            ),
        ],
    )
    def test_construct_component(
        self,
        measurement_type: Type[Measurement],
        measurement_args: Dict[str, Any],
        expectation: ContextManager[None],
    ):
        """Initialization behaves correctly.

        The measurement is constructed correctly and raises exceptions
        when initialized with invalid inputs.

        Args:
            measurement_type: The type of the measurement to be constructed.
            measurement_args: The arguments to the measurement.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        super().test_construct_component(
            measurement_type, measurement_args, expectation, exception_properties=None
        )

    @pytest.mark.parametrize(
        "measurement, input_data",
        [
            (
                AddGaussianNoise(
                    input_domain=NumpyIntegerDomain(), sigma_squared=float("inf")
                ),
                np.int64(1),
            ),
            (
                AddGaussianNoise(
                    input_domain=NumpyFloatDomain(), sigma_squared=float("inf")
                ),
                np.float64(-1),
            ),
        ],
    )
    def test_infinite_protection(self, measurement: Measurement, input_data: Any):
        """Measurement works with infinite protection.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
        """
        output = measurement(input_data)
        assert output in [float("inf"), -float("inf")]

    @pytest.mark.skip("No arguments to mutate")
    def test_mutable_inputs(
        self,
        measurement_type: Type[Measurement],
        measurement_args: Dict[str, Any],
        key: str,
        mutator: Callable,
    ):
        """Mutable inputs to the measurement are copied.

        Args:
            measurement_type: The type of the measurement to be constructed.
            measurement_args: The arguments to the measurement.
            key: The parameter name to be changed.
            mutator: A lambda function that mutates the parameter.
        """
        super().test_mutable_inputs(measurement_type, measurement_args, key, mutator)

    @pytest.mark.parametrize(
        "measurement, input_data, expected_output",
        [
            (
                AddGaussianNoise(input_domain=NumpyIntegerDomain(), sigma_squared=0),
                np.int64(10),
                10.0,
            ),
            (
                AddGaussianNoise(
                    input_domain=NumpyFloatDomain(size=32), sigma_squared=0
                ),
                np.float32(-0.1234567),
                float(np.float32(-0.1234567)),
            ),
        ],
    )
    def test_no_protection(
        self, measurement: Measurement, input_data: Any, expected_output: Any
    ):
        """Measurement works with no protection.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
            expected_output: Expected unprotected output.
        """
        super().test_no_protection(measurement, input_data, expected_output)

    @pytest.mark.parametrize(
        "measurement, input_data, expected_output_type",
        [
            (
                AddGaussianNoise(input_domain=NumpyIntegerDomain(), sigma_squared=1),
                np.int64(10),
                float,
            ),
            (
                AddGaussianNoise(
                    input_domain=NumpyFloatDomain(size=32), sigma_squared=1
                ),
                np.float32(-0.1234567),
                float,
            ),
        ],
    )
    def test_output(
        self, measurement: Measurement, input_data: Any, expected_output_type: type
    ):
        """Measurement produces an output that has the expected type.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
            expected_output_type: The expected type for the output.
        """
        super().test_output(measurement, input_data, expected_output_type)

    @pytest.mark.parametrize(
        "measurement, d_in, expected_d_out, expectation",
        [
            (
                AddGaussianNoise(input_domain=NumpyIntegerDomain(), sigma_squared=1),
                1,
                "0.5",
                does_not_raise(),
            ),
            (
                AddGaussianNoise(input_domain=NumpyIntegerDomain(), sigma_squared=2),
                1,
                "0.25",
                does_not_raise(),
            ),
            (
                AddGaussianNoise(input_domain=NumpyFloatDomain(), sigma_squared="0.5"),
                1,
                1,
                does_not_raise(),
            ),
            (
                AddGaussianNoise(input_domain=NumpyFloatDomain(), sigma_squared="0.5"),
                3,
                9,
                does_not_raise(),
            ),
            (
                AddGaussianNoise(input_domain=NumpyFloatDomain(), sigma_squared="0.5"),
                -1,
                None,
                pytest.raises(
                    ValueError,
                    match="Invalid value for metric AbsoluteDifference: -1 is not "
                    "greater than or equal to 0",
                ),
            ),
            (
                AddGaussianNoise(input_domain=NumpyIntegerDomain(), sigma_squared=0),
                1,
                float("inf"),
                does_not_raise(),
            ),
            (
                AddGaussianNoise(input_domain=NumpyIntegerDomain(), sigma_squared=0),
                0,
                0,
                does_not_raise(),
            ),
            (
                AddGaussianNoise(
                    input_domain=NumpyIntegerDomain(), sigma_squared=float("inf")
                ),
                1,
                0,
                does_not_raise(),
            ),
        ],
    )
    def test_privacy_function(
        self,
        measurement: Measurement,
        d_in: Any,
        expected_d_out: Any,
        expectation: ContextManager[None],
    ):
        """Testing the measurement's privacy function.

        Args:
            measurement: The measurement to be tested.
            d_in: Distance between inputs.
            expected_d_out: Expected distance between outputs.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        super().test_privacy_function(
            measurement, d_in, expected_d_out, expectation, exception_properties=None
        )

    @pytest.mark.parametrize(
        "measurement, d_in, d_out, expected, expectation",
        [
            (
                AddGaussianNoise(input_domain=NumpyIntegerDomain(), sigma_squared=1),
                1,
                "1/2",
                True,
                does_not_raise(),
            ),
            (
                AddGaussianNoise(input_domain=NumpyIntegerDomain(), sigma_squared=1),
                1,
                2,
                True,
                does_not_raise(),
            ),
            (
                AddGaussianNoise(input_domain=NumpyIntegerDomain(), sigma_squared=1),
                1,
                "0.3",
                False,
                does_not_raise(),
            ),
            (
                AddGaussianNoise(input_domain=NumpyFloatDomain(), sigma_squared="0.5"),
                3,
                10,
                True,
                does_not_raise(),
            ),
            (
                AddGaussianNoise(input_domain=NumpyFloatDomain(), sigma_squared="0.5"),
                3,
                5,
                False,
                does_not_raise(),
            ),
            (
                AddGaussianNoise(input_domain=NumpyFloatDomain(), sigma_squared="0.5"),
                -1,
                1,
                True,
                pytest.raises(
                    ValueError,
                    match="Invalid value for metric AbsoluteDifference: -1 is not "
                    "greater than or equal to 0",
                ),
            ),
            (
                AddGaussianNoise(input_domain=NumpyFloatDomain(), sigma_squared="0.5"),
                tuple(),
                1,
                True,
                pytest.raises(
                    TypeError,
                    match=re.escape(
                        'type of argument "d_in" must be one of (ExactNumber, float, '
                        "int, str, Fraction, Expr); got tuple instead"
                    ),
                ),
            ),
            (
                AddGaussianNoise(input_domain=NumpyFloatDomain(), sigma_squared="0.5"),
                1,
                -1,
                True,
                pytest.raises(
                    ValueError,
                    match=re.escape(
                        "Invalid RhoZCDP measure value (rho): -1 is not greater than "
                        "or equal to 0"
                    ),
                ),
            ),
            (
                AddGaussianNoise(input_domain=NumpyFloatDomain(), sigma_squared="0.5"),
                1,
                {},
                True,
                pytest.raises(
                    TypeError,
                    match=re.escape(
                        'type of argument "value" must be one of (ExactNumber, float, '
                        "int, str, Fraction, Expr); got dict instead"
                    ),
                ),
            ),
            (
                AddGaussianNoise(input_domain=NumpyIntegerDomain(), sigma_squared=0),
                1,
                float("inf"),
                True,
                does_not_raise(),
            ),
            (
                AddGaussianNoise(input_domain=NumpyIntegerDomain(), sigma_squared=0),
                0,
                0,
                True,
                does_not_raise(),
            ),
            (
                AddGaussianNoise(
                    input_domain=NumpyIntegerDomain(), sigma_squared=float("inf")
                ),
                1,
                0,
                True,
                does_not_raise(),
            ),
        ],
    )
    def test_privacy_relation(
        self,
        measurement: Measurement,
        d_in: Any,
        d_out: Any,
        expected: bool,
        expectation: ContextManager[None],
    ):
        """Testing the measurement's privacy function.

        Args:
            measurement: The measurement to be tested.
            d_in: Distance between inputs.
            d_out: Distance between outputs.
            expected: Whether the d_in, d_out pair is close.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        super().test_privacy_relation(
            measurement, d_in, d_out, expected, expectation, exception_properties=None
        )

    @pytest.mark.parametrize(
        "measurement, expected_properties",
        [
            (
                AddGaussianNoise(input_domain=NumpyIntegerDomain(), sigma_squared=1),
                {
                    "input_domain": NumpyIntegerDomain(),
                    "sigma_squared": 1,
                    "input_metric": AbsoluteDifference(),
                    "output_type": DoubleType(),
                    "output_measure": RhoZCDP(),
                    "is_interactive": False,
                },
            ),
            (
                AddGaussianNoise(
                    input_domain=NumpyFloatDomain(size=32), sigma_squared=2
                ),
                {
                    "input_domain": NumpyFloatDomain(size=32),
                    "sigma_squared": 2,
                    "input_metric": AbsoluteDifference(),
                    "output_type": DoubleType(),
                    "output_measure": RhoZCDP(),
                    "is_interactive": False,
                },
            ),
        ],
    )
    def test_properties(
        self, measurement: Measurement, expected_properties: Dict[str, Any]
    ):
        """All properties have the expected values.

        Args:
            measurement: The constructed measurement to be tested.
            expected_properties: A dictionary containing all the
                property:value pairs measurement is expected to have.
        """
        super().test_properties(measurement, expected_properties)

    @pytest.mark.parametrize(
        "measurement",
        [AddGaussianNoise(input_domain=NumpyIntegerDomain(), sigma_squared=1)],
    )
    def test_property_immutability(self, measurement: Measurement):
        """The properties return copies for mutable values.

        Args:
            measurement: The measurement to be tested.
        """
        super().test_property_immutability(measurement)


class TestAddDiscreteGaussianNoise(MeasurementTests):
    """:class:`~tmlt.core.measurements.noise_mechanisms.AddDiscreteGaussianNoise`."""

    @pytest.fixture
    def measurement_type(self) -> Type[Measurement]:  # pylint: disable=no-self-use
        """Returns the type of the measurement to be tested."""
        return AddDiscreteGaussianNoise

    @pytest.mark.parametrize(
        "measurement_args, expectation",
        [
            ({"sigma_squared": sp.Rational("0.3")}, does_not_raise()),
            (
                {"sigma_squared": "x"},
                pytest.raises(
                    ValueError, match="Invalid sigma_squared: x contains free symbols"
                ),
            ),
            (
                {"sigma_squared": -1},
                pytest.raises(
                    ValueError,
                    match="Invalid sigma_squared: -1 is not greater than or equal to 0",
                ),
            ),
            (
                {"sigma_squared": float("inf")},
                pytest.raises(
                    ValueError,
                    match="Invalid sigma_squared: inf is not strictly less than inf",
                ),
            ),
            ({"sigma_squared": "0"}, does_not_raise()),
        ],
    )
    def test_construct_component(
        self,
        measurement_type: Type[Measurement],
        measurement_args: Dict[str, Any],
        expectation: ContextManager[None],
    ):
        """Initialization behaves correctly.

        The measurement is constructed correctly and raises exceptions
        when initialized with invalid inputs.

        Args:
            measurement_type: The type of the measurement to be constructed.
            measurement_args: The arguments to the measurement.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        super().test_construct_component(
            measurement_type, measurement_args, expectation, exception_properties=None
        )

    @pytest.mark.skip("Doesn't support infinite protection")
    def test_infinite_protection(
        self, measurement: Measurement, input_data: Any, expected_output: Any
    ):
        """Measurement works with infinite protection.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
            expected_output: The expected output of the measurement.
        """
        super().test_infinite_protection(measurement, input_data, expected_output)

    @pytest.mark.skip("No arguments to mutate")
    def test_mutable_inputs(
        self,
        measurement_type: Type[Measurement],
        measurement_args: Dict[str, Any],
        key: str,
        mutator: Callable,
    ):
        """Mutable inputs to the measurement are copied.

        Args:
            measurement_type: The type of the measurement to be constructed.
            measurement_args: The arguments to the measurement.
            key: The parameter name to be changed.
            mutator: A lambda function that mutates the parameter.
        """
        super().test_mutable_inputs(measurement_type, measurement_args, key, mutator)

    @pytest.mark.parametrize(
        "measurement, input_data, expected_output",
        [
            (AddDiscreteGaussianNoise(sigma_squared=0), np.int64(10), 10),
            (AddDiscreteGaussianNoise(sigma_squared=0), np.int64(-10), -10),
        ],
    )
    def test_no_protection(
        self, measurement: Measurement, input_data: Any, expected_output: Any
    ):
        """Measurement works with no protection.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
            expected_output: Expected unprotected output.
        """
        super().test_no_protection(measurement, input_data, expected_output)

    @pytest.mark.parametrize(
        "measurement, input_data, expected_output_type",
        [(AddDiscreteGaussianNoise(sigma_squared=3), np.int64(0), int)],
    )
    def test_output(
        self, measurement: Measurement, input_data: Any, expected_output_type: type
    ):
        """Measurement produces an output that has the expected type.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
            expected_output_type: The expected type for the output.
        """
        super().test_output(measurement, input_data, expected_output_type)

    @pytest.mark.parametrize(
        "measurement, d_in, expected_d_out, expectation",
        [
            (AddDiscreteGaussianNoise(sigma_squared=1), 1, "1/2", does_not_raise()),
            (AddDiscreteGaussianNoise(sigma_squared=2), 1, "1/4", does_not_raise()),
            (AddDiscreteGaussianNoise(sigma_squared="0.5"), 1, 1, does_not_raise()),
            (AddDiscreteGaussianNoise(sigma_squared="0.5"), 3, 9, does_not_raise()),
            (
                AddDiscreteGaussianNoise(sigma_squared="0.5"),
                -1,
                None,
                pytest.raises(
                    ValueError,
                    match="Invalid value for metric AbsoluteDifference: -1 is not "
                    "greater than or equal to 0",
                ),
            ),
            (
                AddDiscreteGaussianNoise(sigma_squared=0),
                1,
                float("inf"),
                does_not_raise(),
            ),
            (AddDiscreteGaussianNoise(sigma_squared=0), 0, 0, does_not_raise()),
        ],
    )
    def test_privacy_function(
        self,
        measurement: Measurement,
        d_in: Any,
        expected_d_out: Any,
        expectation: ContextManager[None],
    ):
        """Testing the measurement's privacy function.

        Args:
            measurement: The measurement to be tested.
            d_in: Distance between inputs.
            expected_d_out: Expected distance between outputs.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        super().test_privacy_function(
            measurement, d_in, expected_d_out, expectation, exception_properties=None
        )

    @pytest.mark.parametrize(
        "measurement, d_in, d_out, expected, expectation",
        [
            (
                AddDiscreteGaussianNoise(sigma_squared=1),
                1,
                "1/2",
                True,
                does_not_raise(),
            ),
            (AddDiscreteGaussianNoise(sigma_squared=1), 1, 1, True, does_not_raise()),
            (
                AddDiscreteGaussianNoise(sigma_squared=1),
                1,
                "0.3",
                False,
                does_not_raise(),
            ),
            (
                AddDiscreteGaussianNoise(sigma_squared="0.5"),
                3,
                9,
                True,
                does_not_raise(),
            ),
            (
                AddDiscreteGaussianNoise(sigma_squared="0.5"),
                3,
                7,
                False,
                does_not_raise(),
            ),
            (
                AddDiscreteGaussianNoise(sigma_squared="0.5"),
                -1,
                1,
                True,
                pytest.raises(
                    ValueError,
                    match="Invalid value for metric AbsoluteDifference: -1 is not "
                    "greater than or equal to 0",
                ),
            ),
            (
                AddDiscreteGaussianNoise(sigma_squared="0.5"),
                tuple(),
                1,
                True,
                pytest.raises(
                    TypeError,
                    match=re.escape(
                        'type of argument "d_in" must be one of (ExactNumber, float, '
                        "int, str, Fraction, Expr); got tuple instead"
                    ),
                ),
            ),
            (
                AddDiscreteGaussianNoise(sigma_squared="0.5"),
                1,
                -1,
                True,
                pytest.raises(
                    ValueError,
                    match=re.escape(
                        "Invalid RhoZCDP measure value (rho): -1 is not greater than "
                        "or equal to 0"
                    ),
                ),
            ),
            (
                AddDiscreteGaussianNoise(sigma_squared="0.5"),
                1,
                {},
                True,
                pytest.raises(
                    TypeError,
                    match=re.escape(
                        'type of argument "value" must be one of (ExactNumber, float, '
                        "int, str, Fraction, Expr); got dict instead"
                    ),
                ),
            ),
            (
                AddDiscreteGaussianNoise(sigma_squared=0),
                1,
                float("inf"),
                True,
                does_not_raise(),
            ),
            (AddDiscreteGaussianNoise(sigma_squared=0), 0, 0, True, does_not_raise()),
        ],
    )
    def test_privacy_relation(
        self,
        measurement: Measurement,
        d_in: Any,
        d_out: Any,
        expected: bool,
        expectation: ContextManager[None],
    ):
        """Testing the measurement's privacy function.

        Args:
            measurement: The measurement to be tested.
            d_in: Distance between inputs.
            d_out: Distance between outputs.
            expected: Whether the d_in, d_out pair is close.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        super().test_privacy_relation(
            measurement, d_in, d_out, expected, expectation, exception_properties=None
        )

    @pytest.mark.parametrize(
        "measurement, expected_properties",
        [
            (
                AddDiscreteGaussianNoise(sigma_squared=1),
                {
                    "input_domain": NumpyIntegerDomain(),
                    "sigma_squared": 1,
                    "input_metric": AbsoluteDifference(),
                    "output_type": LongType(),
                    "output_measure": RhoZCDP(),
                    "is_interactive": False,
                },
            ),
            (
                AddDiscreteGaussianNoise(sigma_squared=2),
                {
                    "input_domain": NumpyIntegerDomain(),
                    "sigma_squared": 2,
                    "input_metric": AbsoluteDifference(),
                    "output_type": LongType(),
                    "output_measure": RhoZCDP(),
                    "is_interactive": False,
                },
            ),
        ],
    )
    def test_properties(
        self, measurement: Measurement, expected_properties: Dict[str, Any]
    ):
        """All properties have the expected values.

        Args:
            measurement: The constructed measurement to be tested.
            expected_properties: A dictionary containing all the
                property:value pairs measurement is expected to have.
        """
        super().test_properties(measurement, expected_properties)

    @pytest.mark.parametrize("measurement", [AddDiscreteGaussianNoise(sigma_squared=1)])
    def test_property_immutability(self, measurement: Measurement):
        """The properties return copies for mutable values.

        Args:
            measurement: The measurement to be tested.
        """
        super().test_property_immutability(measurement)

    @pytest.mark.parametrize(
        "sigma_squared, p, expected, expectation",
        [
            (1, 0.5, 0, does_not_raise()),
            (2, 0.5, 0, does_not_raise()),
            (3.1415, 0.5, 0, does_not_raise()),
            (1, 0.75, 1, does_not_raise()),
            (1, 0.25, -1, does_not_raise()),
            (3.5, 0.86, 2, does_not_raise()),
            (3.5, 0.14, -2, does_not_raise()),
            (0, 0.6, 0, does_not_raise()),  # 0 scale is 0
            (  # inf scale is +inf 50% of the time
                float("inf"),
                0.6,
                float("inf"),
                does_not_raise(),
            ),
            (  # inf scale is -inf 50% of the time
                float("inf"),
                0.4,
                float("-inf"),
                does_not_raise(),
            ),
            (1, 0, float("-inf"), does_not_raise()),  # p=0 is -inf
            (1, 1, float("inf"), does_not_raise()),  # p=1 is +inf
            (-1, 0.6, float("nan"), does_not_raise()),  # negative scale is nan
            (float("nan"), 0.6, float("nan"), does_not_raise()),  # nan scale is nan
        ]
        + [  # p values outside [0, 1] raise ValueError
            (
                1,
                bad_p,
                float("nan"),
                pytest.raises(
                    ValueError,
                    match=re.escape(
                        "Probabilities input to the inverse CDF must be in [0, 1], "
                        f"but got {bad_p}"
                    ),
                ),
            )
            for bad_p in [-1, 1.1, float("inf"), float("nan")]
        ],
    )
    def test_inverse_cdf(
        self,
        sigma_squared: float,
        p: float,
        expected: float,
        expectation: ContextManager[None],
    ):
        """Testing the inverse_cdf function.

        Args:
            sigma_squared: The sigma_squared of the DiscreteGaussian distribution.
            p: The probability.
            expected: The expected value.
            expectation: A context manager that captures the correct expected type of
        """
        with expectation:
            actual = AddDiscreteGaussianNoise.inverse_cdf(
                sigma_squared=sigma_squared, probability=p
            )
            if np.isnan(expected):
                assert np.isnan(actual), f"Expected {expected}, got {actual}"
            else:
                assert actual == pytest.approx(
                    expected
                ), f"Expected {expected}, got {actual}"



"""Abstract class for testing input/output measurements."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2022

import copy

# pylint: disable=unused-import
from abc import ABC, abstractmethod
from contextlib import nullcontext as does_not_raise
from typing import Any, Callable, ContextManager, Dict, Optional, Type

import pytest

from tmlt.core.measurements.base import Measurement
from tmlt.core.utils.testing import assert_property_immutability, get_all_props


class MeasurementTests(ABC):
    """Test classes for Measurements should inherit from this class."""

    @abstractmethod
    @pytest.fixture
    def measurement_type(self) -> Type[Measurement]:  # pylint: disable=no-self-use
        """Returns the type of the measurement to be tested."""
        return Measurement

    @abstractmethod
    def test_mutable_inputs(
        self,
        measurement_type: Type[Measurement],
        measurement_args: Dict[str, Any],
        key: str,
        mutator: Callable,
    ):
        """Mutable inputs to the measurement are copied.

        Args:
            measurement_type: The type of the measurement to be constructed.
            measurement_args: The arguments to the measurement.
            key: The parameter name to be changed.
            mutator: A lambda function that mutates the parameter.
        """
        measurement = measurement_type(**measurement_args)
        assert hasattr(measurement, key), f"{key} not in {measurement}"
        expected = copy.deepcopy(getattr(measurement, key))
        mutator(measurement_args[key])
        actual_value = getattr(measurement, key)
        assert (
            getattr(measurement, key) == expected
        ), f"Expected {key} to be {expected}, got {actual_value}"

    @abstractmethod
    def test_property_immutability(self, measurement: Measurement):
        """The properties return copies for mutable values.

        Args:
            measurement: The measurement to be tested.
        """
        props = [prop[0] for prop in get_all_props(type(measurement))]
        for prop in props:
            assert_property_immutability(measurement, prop)

    @abstractmethod
    def test_properties(
        self, measurement: Measurement, expected_properties: Dict[str, Any]
    ):
        """All properties have the expected values.

        Args:
            measurement: The constructed measurement to be tested.
            expected_properties: A dictionary containing all the
                property:value pairs measurement is expected to have.
        """
        actual_props = [prop[0] for prop in get_all_props(type(measurement))]
        assert set(expected_properties.keys()) == set(actual_props), (
            "Keys of expected properties don't match actual properties. Expected "
            f"{set(expected_properties.keys())}; got {set(actual_props)}"
        )
        for prop, expected_val in expected_properties.items():
            assert hasattr(measurement, prop), f"{prop} not in {measurement}"
            actual_value = getattr(measurement, prop)
            assert (
                getattr(measurement, prop) == expected_val
            ), f"Expected {prop} to be {expected_val}; got {actual_value}"

    @abstractmethod
    def test_construct_component(
        self,
        measurement_type: Type[Measurement],
        measurement_args: Dict[str, Any],
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Initialization behaves correctly.

        The measurement is constructed correctly and raises exceptions
        when initialized with invalid inputs.

        Args:
            measurement_type: The type of the measurement to be constructed.
            measurement_args: The arguments to the measurement.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        with expectation as exception:
            measurement_type(**measurement_args)
        if exception_properties is None or len(exception_properties) == 0:
            return
        # Help out mypy
        assert isinstance(exception, pytest.ExceptionInfo)
        for prop, expected_value in exception_properties.items():
            assert hasattr(exception.value, prop), f"{prop} not in {exception.value}"
            actual_value = getattr(exception.value, prop)
            assert (
                actual_value == expected_value
            ), f"Expected {prop} to be {expected_value}, got {actual_value}"

    @abstractmethod
    def test_privacy_function(
        self,
        measurement: Measurement,
        d_in: Any,
        expected_d_out: Any,
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Testing the measurement's privacy function.

        Args:
            measurement: The measurement to be tested.
            d_in: Distance between inputs.
            expected_d_out: Expected distance between outputs.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        with expectation as exception:
            d_out = measurement.privacy_function(d_in)
            assert d_out == expected_d_out, f"Expected {expected_d_out}, got {d_out}"
        if exception_properties is None or len(exception_properties) == 0:
            return
        for prop, expected_value in exception_properties.items():
            assert hasattr(exception, prop), f"{prop} not in {exception}"
            actual_value = getattr(exception, prop)
            assert (
                getattr(exception, prop) == expected_value
            ), f"Expected {prop} to be {expected_value}, got {actual_value}"

    @abstractmethod
    def test_privacy_relation(
        self,
        measurement: Measurement,
        d_in: Any,
        d_out: Any,
        expected: bool,
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Testing the measurement's privacy function.

        Args:
            measurement: The measurement to be tested.
            d_in: Distance between inputs.
            d_out: Distance between outputs.
            expected: Whether the d_in, d_out pair is close.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        with expectation as exception:
            assert measurement.privacy_relation(d_in, d_out) == expected
        if exception_properties is None or len(exception_properties) == 0:
            return
        for prop, expected_value in exception_properties.items():
            assert hasattr(exception, prop), f"{prop} not in {exception}"
            actual_value = getattr(exception, prop)
            assert (
                getattr(exception, prop) == expected_value
            ), f"Expected {prop} to be {expected_value}, got {actual_value}"

    @abstractmethod
    def test_output(
        self, measurement: Measurement, input_data: Any, expected_output_type: type
    ):
        """Measurement produces an output that has the expected type.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
            expected_output_type: The expected type for the output.
        """
        output = measurement(input_data)
        assert isinstance(output, expected_output_type)

    @abstractmethod
    def test_no_protection(
        self, measurement: Measurement, input_data: Any, expected_output: Any
    ):
        """Measurement works with no protection.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
            expected_output: Expected unprotected output.
        """
        output = measurement(input_data)
        assert output == expected_output

    @abstractmethod
    def test_infinite_protection(
        self, measurement: Measurement, input_data: Any, expected_output: Any
    ):
        """Measurement works with infinite protection.

        Args:
            measurement: The measurement.
            input_data: The input data for the measurement.
            expected_output: Expected infinitely protected output.
        """
        output = measurement(input_data)
        assert output == expected_output



"""Utilities related to joining dataframes."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

import dataclasses
from typing import Dict, List, Optional, Tuple, Union

from pyspark.sql import DataFrame
from pyspark.sql import functions as sf
from typeguard import typechecked

from tmlt.core.domains.base import Domain
from tmlt.core.domains.spark_domains import (
    SparkColumnDescriptor,
    SparkDataFrameDomain,
    SparkDateColumnDescriptor,
    SparkFloatColumnDescriptor,
    SparkIntegerColumnDescriptor,
    SparkStringColumnDescriptor,
    SparkTimestampColumnDescriptor,
)
from tmlt.core.utils.misc import get_nonconflicting_string

# pylint: disable=no-member


def natural_join_columns(
    left_columns: List[str], right_columns: List[str]
) -> List[str]:
    """Returns the columns to join on to perform a natural join.

    The columns to join on are the ones that are in both `left_columns` and
    `right_columns`, in the order they appear in `left_columns`.

    .. note::

        Returns an empty list if there are no common columns.

    Examples:
        >>> natural_join_columns(["a", "b", "c"], ["b", "c", "d"])
        ['b', 'c']
        >>> natural_join_columns(["a", "b", "c"], ["d", "e", "f"])
        []

    Args:
        left_columns: Columns of the left dataframe.
        right_columns: Columns of the right dataframe.
    """
    return [column for column in left_columns if column in right_columns]


@typechecked
def columns_after_join(
    left_columns: List[str], right_columns: List[str], on: Optional[List[str]] = None
) -> Dict[str, Union[Tuple[str, str], Tuple[str, None], Tuple[None, str]]]:
    """Return the expected output columns and their origin from joining two dataframes.

    The keys are the names of the output columns in the order they will appear in the
    output dataframe. The values are tuples of the form (left_column, right_column),
    where left_column and right_column are the names of the columns in the left and
    right dataframes that the output column is derived from. If the output column is
    not derived from a column in the left or right dataframe, the corresponding value
    is None.

    The output columns are ordered as follows:

        - Join columns (in the order given by the user, or the order they are in the
          left table if not provided) appear first.
        - Columns of left table (with _left appended as required) appear
          next in the input order. (excluding join columns)
        - Columns of the right table (with _right appended as required) appear
          last in the input order. (excluding join columns)

    _left and _right are appended to output column names if they are in both
    `left_columns` and `right_columns`, but are not joined on.

    Also does input validation. Checks:

        - The join involves at least one column.
        - Join columns are in both tables.
        - None of the column names are duplicated in any of the inputs.
        - No name collisions when adding _left or _right to a column name.

    Examples:
        >>> columns_after_join(["a", "b", "c"], ["b", "c", "d"])
        {'b': ('b', 'b'), 'c': ('c', 'c'), 'a': ('a', None), 'd': (None, 'd')}
        >>> columns_after_join(["a", "b", "c"], ["b", "c", "d"], ["b"])
        {'b': ('b', 'b'), 'a': ('a', None), 'c_left': ('c', None), 'c_right': (None, 'c'), 'd': (None, 'd')}
        >>> columns_after_join(["a", "b", "c"], ["d", "e", "f"])
        Traceback (most recent call last):
        ...
        ValueError: Join must involve at least one column.
        >>> columns_after_join(["a", "b", "c"], ["a", "b", "c"], ["d"])
        Traceback (most recent call last):
        ...
        ValueError: Join column 'd' not in the left table.
        >>> columns_after_join(["a", "b", "c"], ["a", "b", "c"], ["a", "a"])
        Traceback (most recent call last):
        ...
        ValueError: Join columns (`on`) contain duplicates.
        >>> columns_after_join(["a", "b"], ["a", "b", "b_right"], ["a"])
        Traceback (most recent call last):
        ...
        ValueError: Name collision, 'b_right' would appear more than once in the output.


    Args:
        left_columns: Columns of the left dataframe.
        right_columns: Columns of the right dataframe.
        on: Columns to join on. If None, join on all columns with the same
            name.
    """  # pylint: disable=line-too-long
    if on is None:
        on = natural_join_columns(left_columns, right_columns)
    if len(on) == 0:
        raise ValueError("Join must involve at least one column.")
    for column in on:
        if column not in left_columns:
            raise ValueError(f"Join column '{column}' not in the left table.")
        if column not in right_columns:
            raise ValueError(f"Join column '{column}' not in the right table.")
    if len(set(on)) != len(on):
        raise ValueError("Join columns (`on`) contain duplicates.")
    if len(set(left_columns)) != len(left_columns):
        raise ValueError("Left columns contain duplicates.")
    if len(set(right_columns)) != len(right_columns):
        raise ValueError("Right columns contain duplicates.")

    output_columns: Dict[
        str, Union[Tuple[str, str], Tuple[str, None], Tuple[None, str]]
    ] = {column: (column, column) for column in on}
    for column in left_columns:
        if column in on:
            continue
        if column in right_columns:
            new_column = f"{column}_left"
        else:
            new_column = column
        if new_column in output_columns:
            raise ValueError(
                f"Name collision, '{new_column}' would appear more than once in the"
                " output."
            )
        output_columns[new_column] = (column, None)
    for column in right_columns:
        if column in on:
            continue
        if column in left_columns:
            new_column = f"{column}_right"
        else:
            new_column = column
        if new_column in output_columns:
            raise ValueError(
                f"Name collision, '{new_column}' would appear more than once in the"
                " output."
            )
        output_columns[new_column] = (None, column)
    return output_columns


@typechecked
def domain_after_join(
    left_domain: Domain,
    right_domain: Domain,
    on: Optional[List[str]] = None,
    how: str = "inner",
    nulls_are_equal: bool = False,
) -> SparkDataFrameDomain:
    """Returns the domain of the join of two dataframes.

    Also does input validation. Checks:

        - All checks from :func:`~.columns_after_join`.
        - `how` is one of "left", "right", "inner", or "outer".
        - Join columns have the same data type.
        - Left and right domains are SparkDataFrameDomains.

    .. note::

        This takes into account extra metadata about the columns, such as whether nulls/
        infs are allowed, and what kind of join is performed.

        See :ref:`NaNs, nulls, and infs <special-values>` for more information about
        comparisons involving special values.

    Args:
        left_domain: Domain of the left dataframe.
        right_domain: Domain of the right dataframe.
        on: Columns to join on. If None, join on all columns with the same
            name.
        how: Join type. Must be one of "left", "right", "inner", "outer". This
            defaults to "inner".
        nulls_are_equal: If True, treats null values as equal. Defaults to False.
    """
    if not isinstance(left_domain, SparkDataFrameDomain):
        raise TypeError("Left join input domain must be a SparkDataFrameDomain.")
    if not isinstance(right_domain, SparkDataFrameDomain):
        raise TypeError("Right join input domain must be a SparkDataFrameDomain.")
    if on is None:
        on = natural_join_columns(
            left_columns=list(left_domain.schema),
            right_columns=list(right_domain.schema),
        )
    if how not in ["left", "right", "inner", "outer"]:
        raise ValueError(
            "Join type (`how`) must be one of 'left', 'right', 'inner', or 'outer', not"
            f" '{how}'."
        )
    output_columns = columns_after_join(
        left_columns=list(left_domain.schema),
        right_columns=list(right_domain.schema),
        on=on,
    )
    output_descriptors: Dict[str, SparkColumnDescriptor] = {}
    for output_column, (left_column, right_column) in output_columns.items():
        left_descriptor = left_domain.schema.get(left_column, None)  # type: ignore
        right_descriptor = right_domain.schema.get(right_column, None)  # type: ignore
        if left_descriptor is None:
            assert right_descriptor is not None
            output_descriptors[output_column] = dataclasses.replace(  # type: ignore
                right_descriptor,  # type: ignore
                allow_null=right_descriptor.allow_null or how in ["left", "outer"],
            )
            continue
        if right_descriptor is None:
            assert left_descriptor is not None
            output_descriptors[output_column] = dataclasses.replace(  # type: ignore
                left_descriptor,  # type: ignore
                allow_null=left_descriptor.allow_null or how in ["right", "outer"],
            )
            continue
        assert left_descriptor is not None
        assert right_descriptor is not None
        # The only remaining case is when the output column is a join column.
        assert output_column in on
        if left_descriptor.data_type != right_descriptor.data_type:
            # str(left_descriptor.datatype) changes based on the version of Spark
            left_dtype = str(left_descriptor.data_type).replace("()", "")
            right_dtype = str(right_descriptor.data_type).replace("()", "")
            raise ValueError(
                f"'{output_column}' has different data types in left "
                f"({left_dtype}) and right "
                f"({right_dtype}) domains."
            )
        # All column types are nullable
        allow_null = None
        if how == "left":
            allow_null = left_descriptor.allow_null
        elif how == "right":
            allow_null = right_descriptor.allow_null
        elif how == "inner":
            allow_null = (
                left_descriptor.allow_null and right_descriptor.allow_null
                if nulls_are_equal
                else False
            )
        elif how == "outer":
            allow_null = left_descriptor.allow_null or right_descriptor.allow_null
        assert allow_null is not None
        new_descriptor: SparkColumnDescriptor
        if isinstance(left_descriptor, SparkIntegerColumnDescriptor):
            assert isinstance(right_descriptor, SparkIntegerColumnDescriptor)
            assert left_descriptor.size == right_descriptor.size
            new_descriptor = SparkIntegerColumnDescriptor(
                allow_null=allow_null, size=left_descriptor.size
            )
        elif isinstance(left_descriptor, SparkFloatColumnDescriptor):
            assert isinstance(right_descriptor, SparkFloatColumnDescriptor)
            allow_nan = None
            allow_inf = None
            if how == "left":
                allow_nan = left_descriptor.allow_nan
                allow_inf = left_descriptor.allow_inf
            elif how == "right":
                allow_nan = right_descriptor.allow_nan
                allow_inf = right_descriptor.allow_inf
            elif how == "inner":
                allow_nan = left_descriptor.allow_nan and right_descriptor.allow_nan
                allow_inf = left_descriptor.allow_inf and right_descriptor.allow_inf
            elif how == "outer":
                allow_nan = left_descriptor.allow_nan or right_descriptor.allow_nan
                allow_inf = left_descriptor.allow_inf or right_descriptor.allow_inf
            assert allow_nan is not None
            assert allow_inf is not None
            assert left_descriptor.size == right_descriptor.size
            new_descriptor = SparkFloatColumnDescriptor(
                allow_nan=allow_nan,
                allow_inf=allow_inf,
                allow_null=allow_null,
                size=left_descriptor.size,
            )
        elif isinstance(
            left_descriptor,
            (
                SparkStringColumnDescriptor,
                SparkDateColumnDescriptor,
                SparkTimestampColumnDescriptor,
            ),
        ):
            descriptor_class = left_descriptor.__class__
            assert isinstance(right_descriptor, descriptor_class)
            new_descriptor = descriptor_class(allow_null=allow_null)
        else:
            raise NotImplementedError(
                f"Unsupported column descriptor {left_descriptor}."
            )
        output_descriptors[output_column] = new_descriptor
    return SparkDataFrameDomain(output_descriptors)


def join(
    left: DataFrame,
    right: DataFrame,
    on: Optional[List[str]] = None,
    how: str = "inner",
    nulls_are_equal: bool = False,
) -> DataFrame:
    """Returns the join of two dataframes.

    Args:
        left: Left dataframe.
        right: Right dataframe.
        on: Columns to join on. If None, join on all columns with the same
            name.
        how: Join type. Must be one of "left", "right", "inner", "outer". If None,
            defaults to "inner".
        nulls_are_equal: If True, treats null values as equal. Defaults to False.
    """
    # `columns_after_join` and `domain_after_join` are only called for validation.
    columns_after_join(left_columns=left.columns, right_columns=right.columns, on=on)
    domain_after_join(
        left_domain=SparkDataFrameDomain.from_spark_schema(left.schema),
        right_domain=SparkDataFrameDomain.from_spark_schema(right.schema),
        on=on,
        how=how,
    )
    if nulls_are_equal:
        return _join_where_nulls_are_equal(left=left, right=right, on=on, how=how)
    return _join_where_nulls_are_not_equal(left=left, right=right, on=on, how=how)


def _rename_columns(
    left: DataFrame, right: DataFrame, on: Optional[List[str]] = None
) -> Tuple[
    DataFrame,
    DataFrame,
    Dict[str, Union[Tuple[str, str], Tuple[str, None], Tuple[None, str]]],
]:
    """Rename columns in left and right dataframes to avoid conflicts.

    Example:
        ..
            >>> from pyspark.sql import SparkSession
            >>> from tmlt.core.utils.misc import print_sdf
            >>> spark = SparkSession.builder.getOrCreate()
            >>> left = spark.createDataFrame(
            ...     [
            ...         (1, 2, 3),
            ...         (4, 5, 6),
            ...     ],
            ...     ["a", "b", "c"],
            ... )
            >>> right = spark.createDataFrame(
            ...     [
            ...         (2, 7, 8),
            ...         (5, 9, 10),
            ...     ],
            ...     ["b", "c", "d"],
            ... )

        >>> print_sdf(left)
           a  b  c
        0  1  2  3
        1  4  5  6
        >>> print_sdf(right)
           b  c   d
        0  2  7   8
        1  5  9  10
        >>> left_renamed, right_renamed, output_columns = _rename_columns(
        ...     left, right, on=["b"]
        ... )
        >>> print_sdf(left_renamed)
           a  b  c_left
        0  1  2       3
        1  4  5       6
        >>> print_sdf(right_renamed)
           b  c_right   d
        0  2        7   8
        1  5        9  10
        >>> output_columns
        {'b': ('b', 'b'), 'a': ('a', None), 'c_left': ('c', None), 'c_right': (None, 'c'), 'd': (None, 'd')}

    Args:
        left: Left dataframe.
        right: Right dataframe.
        on: Columns to join on. If None, join on all columns with the same
            name.

    Returns:
        A tuple containing

            * Left dataframe with renamed columns.
            * Right dataframe with renamed columns.
            * Mapping from output column name to
              (left column name, right column name). See :func:`columns_after_join`.
    """  # pylint: disable=line-too-long
    if on is None:
        on = natural_join_columns(
            left_columns=left.columns, right_columns=right.columns
        )
    output_columns = columns_after_join(
        left_columns=left.columns, right_columns=right.columns, on=on
    )
    for output_column, (left_column, right_column) in output_columns.items():
        if left_column is not None and left_column != output_column:
            left = left.withColumnRenamed(left_column, output_column)
        if right_column is not None and right_column != output_column:
            right = right.withColumnRenamed(right_column, output_column)
    return left, right, output_columns


def _join_where_nulls_are_not_equal(
    left: DataFrame,
    right: DataFrame,
    on: Optional[List[str]] = None,
    how: str = "inner",
) -> DataFrame:
    """Returns the join of two dataframes, with null values not being equal.

    Args:
        left: Left dataframe.
        right: Right dataframe.
        on: Columns to join on. If None, join on all columns with the same name.
        how: Join type. Must be one of "left", "right", "inner", "outer". If None,
            defaults to "inner".
    """
    left, right, output_columns = _rename_columns(left=left, right=right, on=on)
    return left.join(right, on=on, how=how).select(list(output_columns))


def _join_where_nulls_are_equal(
    left: DataFrame,
    right: DataFrame,
    on: Optional[List[str]] = None,
    how: str = "inner",
) -> DataFrame:
    """Returns the join of two dataframes, with null values being equal.

    Args:
        left: Left dataframe.
        right: Right dataframe.
        on: Columns to join on. If None, join on all columns with the same name.
        how: Join type. Must be one of "left", "right", "inner", "outer". If None,
            defaults to "inner".
    """
    left, right, output_columns = _rename_columns(left=left, right=right, on=on)
    if on is None:
        on = natural_join_columns(
            left_columns=left.columns, right_columns=right.columns
        )
    # Rename left and right columns to avoid confusing Spark
    left_temporary_names: Dict[str, str] = {}
    right_temporary_names: Dict[str, str] = {}
    for column in on:
        left_temporary_names[column] = get_nonconflicting_string(
            left.columns
            + right.columns
            + list(left_temporary_names)
            + list(right_temporary_names)
        )
        right_temporary_names[column] = get_nonconflicting_string(
            left.columns
            + right.columns
            + list(left_temporary_names)
            + list(right_temporary_names)
        )
    left = left.select(
        [
            sf.col(column).alias(left_temporary_names[column])
            if column in on
            else sf.col(column)
            for column in left.columns
        ]
    )
    right = right.select(
        [
            sf.col(column).alias(right_temporary_names[column])
            if column in on
            else sf.col(column)
            for column in right.columns
        ]
    )
    # Join
    condition = [
        left[left_temporary_names[column]].eqNullSafe(
            right[right_temporary_names[column]]
        )
        for column in on
    ]
    result = left.join(right, on=condition, how=how)
    # Merge columns to use original names
    # Consider joining the following inputs:
    # left: {"A": [1]}
    # right: {"A": [2]}
    # using condition = [left["A"].eqNullSafe(right["A"])]
    # and how = "outer"
    # The result so far will be:
    # {"A": [1, None], "A(2)": [None, 2]}
    # Notice that neither of these columns are [1, 2] which is what we eventually want.
    # We need to merge the columns to get the desired result.
    for column in on:
        left_column = left_temporary_names[column]
        right_column = right_temporary_names[column]
        merged_column = (
            sf.when(sf.col(left_column).isNotNull(), sf.col(left_column))
            .when(sf.col(right_column).isNotNull(), sf.col(right_column))
            .otherwise(sf.lit(None))
        )
        result = result.withColumn(column, merged_column)
    # Drop temporary columns and fix order
    return result.select(list(output_columns))



"""Unit tests for :mod:`~tmlt.core.measurements.aggregations`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023
import functools
import unittest
from typing import Callable, List, Optional, Tuple, Union, cast

import sympy as sp
from parameterized import parameterized, parameterized_class
from pyspark.sql import DataFrame
from pyspark.sql.types import StringType, StructField, StructType

from tmlt.core.domains.spark_domains import (
    SparkDataFrameDomain,
    SparkIntegerColumnDescriptor,
    SparkStringColumnDescriptor,
)
from tmlt.core.measurements.aggregations import (
    NoiseMechanism,
    create_average_measurement,
    create_bound_selection_measurement,
    create_count_distinct_measurement,
    create_count_measurement,
    create_partition_selection_measurement,
    create_quantile_measurement,
    create_standard_deviation_measurement,
    create_sum_measurement,
    create_variance_measurement,
)
from tmlt.core.measurements.converters import PureDPToApproxDP, PureDPToRhoZCDP
from tmlt.core.measurements.spark_measurements import BoundSelection
from tmlt.core.measures import (
    ApproxDP,
    PrivacyBudget,
    PrivacyBudgetInput,
    PureDP,
    RhoZCDP,
)
from tmlt.core.metrics import (
    HammingDistance,
    IfGroupedBy,
    RootSumOfSquared,
    SumOf,
    SymmetricDifference,
)
from tmlt.core.transformations.spark_transformations.groupby import GroupBy
from tmlt.core.utils.distributions import double_sided_geometric_cmf_exact
from tmlt.core.utils.exact_number import ExactNumber, ExactNumberInput
from tmlt.core.utils.testing import PySparkTest


@parameterized_class(
    [
        {"group_keys_list": [], "struct_fields": [], "groupby_columns": []},
        {
            "group_keys_list": [("x1",), ("x2",), ("x3",), (None,)],
            "struct_fields": [StructField("A", StringType())],
            "groupby_columns": ["A"],
        },
    ]
)
class TestGroupByAggregationMeasurements(PySparkTest):
    """Tests for :mod:`tmlt.core.measurements.aggregations`."""

    group_keys_list: List[Tuple[str, ...]]
    struct_fields: List[StructField]
    groupby_columns: List[str]

    def setUp(self):
        """Test setup."""
        self.input_domain = SparkDataFrameDomain(
            {
                "A": SparkStringColumnDescriptor(allow_null=True),
                "B": SparkIntegerColumnDescriptor(),
            }
        )
        self.group_keys = self.spark.createDataFrame(
            self.group_keys_list, schema=StructType(self.struct_fields.copy())
        )
        self.sdf = self.spark.createDataFrame(
            [("x1", 2), ("x1", 2), ("x2", 4)], schema=["A", "B"]
        )

    @parameterized.expand(
        [
            (
                input_metric,
                groupby_output_metric,
                output_measure,
                d_out,
                noise_mechanism,
            )
            for noise_mechanism, groupby_output_metric in [
                (NoiseMechanism.LAPLACE, SumOf(SymmetricDifference())),
                (NoiseMechanism.GEOMETRIC, SumOf(SymmetricDifference())),
                (
                    NoiseMechanism.DISCRETE_GAUSSIAN,
                    RootSumOfSquared(SymmetricDifference()),
                ),
            ]
            for input_metric in [
                SymmetricDifference(),
                HammingDistance(),
                IfGroupedBy(
                    "A", cast(Union[SumOf, RootSumOfSquared], groupby_output_metric)
                ),
            ]
            for output_measure, d_out in [
                (PureDP(), sp.Integer(2)),
                (RhoZCDP(), sp.Integer(2)),
                (ApproxDP(), (sp.Integer(2), sp.Integer(0))),
            ]
            if not (
                noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN
                and output_measure != RhoZCDP()
            )
        ]
    )
    def test_create_count_measurement_with_groupby(
        self,
        input_metric: Union[SymmetricDifference, HammingDistance, IfGroupedBy],
        groupby_output_metric: Union[SumOf, RootSumOfSquared],
        output_measure: Union[PureDP, RhoZCDP, ApproxDP],
        d_out: PrivacyBudgetInput,
        noise_mechanism: NoiseMechanism,
    ):
        """Tests that create_count_measurement works correctly with groupby."""
        if self.groupby_columns == [] and isinstance(input_metric, IfGroupedBy):
            return
        count_measurement = create_count_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            output_measure=output_measure,
            noise_mechanism=noise_mechanism,
            d_in=sp.Integer(1),
            d_out=d_out,
            groupby_transformation=GroupBy(
                input_domain=self.input_domain,
                input_metric=input_metric,
                use_l2=isinstance(groupby_output_metric, RootSumOfSquared),
                group_keys=self.group_keys,
            ),
            count_column="test_count",
        )
        self.assertEqual(count_measurement.input_domain, self.input_domain)
        self.assertEqual(count_measurement.output_measure, output_measure)
        self.assertEqual(count_measurement.privacy_function(sp.Integer(1)), d_out)
        answer = count_measurement(self.sdf)
        self.assertIsInstance(answer, DataFrame)
        self.assertEqual(answer.columns, self.groupby_columns + ["test_count"])

    @parameterized.expand(
        [
            (
                input_metric,
                groupby_output_metric,
                output_measure,
                d_out,
                noise_mechanism,
            )
            for noise_mechanism, groupby_output_metric in [
                (NoiseMechanism.LAPLACE, SumOf(SymmetricDifference())),
                (NoiseMechanism.GEOMETRIC, SumOf(SymmetricDifference())),
                (
                    NoiseMechanism.DISCRETE_GAUSSIAN,
                    RootSumOfSquared(SymmetricDifference()),
                ),
            ]
            for input_metric in [
                SymmetricDifference(),
                HammingDistance(),
                IfGroupedBy(
                    "A", cast(Union[SumOf, RootSumOfSquared], groupby_output_metric)
                ),
            ]
            for output_measure, d_out in [
                (PureDP(), sp.Integer(2)),
                (RhoZCDP(), sp.Integer(2)),
                (ApproxDP(), (sp.Integer(2), sp.Integer(0))),
            ]
            if not (
                noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN
                and output_measure != RhoZCDP()
            )
        ]
    )
    def test_create_count_distinct_measurement_with_groupby(
        self,
        input_metric: Union[SymmetricDifference, HammingDistance, IfGroupedBy],
        groupby_output_metric: Union[SumOf, RootSumOfSquared],
        output_measure: Union[PureDP, RhoZCDP, ApproxDP],
        d_out: PrivacyBudgetInput,
        noise_mechanism: NoiseMechanism,
    ):
        """Tests that create_count_distinct_measurement works correctly with groupby."""
        if self.groupby_columns == [] and isinstance(input_metric, IfGroupedBy):
            return
        count_distinct_measurement = create_count_distinct_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            output_measure=output_measure,
            noise_mechanism=noise_mechanism,
            d_in=sp.Integer(1),
            d_out=d_out,
            groupby_transformation=GroupBy(
                input_domain=self.input_domain,
                input_metric=input_metric,
                use_l2=isinstance(groupby_output_metric, RootSumOfSquared),
                group_keys=self.group_keys,
            ),
            count_column="test_count",
        )
        self.assertEqual(count_distinct_measurement.input_domain, self.input_domain)
        self.assertEqual(count_distinct_measurement.output_measure, output_measure)
        self.assertEqual(
            count_distinct_measurement.privacy_function(sp.Integer(1)), d_out
        )
        answer = count_distinct_measurement(self.sdf)
        self.assertIsInstance(answer, DataFrame)
        self.assertEqual(answer.columns, self.groupby_columns + ["test_count"])

    @parameterized.expand(
        [
            (
                input_metric,
                groupby_output_metric,
                output_measure,
                d_out,
                noise_mechanism,
            )
            for noise_mechanism, groupby_output_metric in [
                (NoiseMechanism.LAPLACE, SumOf(SymmetricDifference())),
                (NoiseMechanism.GEOMETRIC, SumOf(SymmetricDifference())),
                (
                    NoiseMechanism.DISCRETE_GAUSSIAN,
                    RootSumOfSquared(SymmetricDifference()),
                ),
            ]
            for input_metric in [
                SymmetricDifference(),
                HammingDistance(),
                IfGroupedBy(
                    "A", cast(Union[SumOf, RootSumOfSquared], groupby_output_metric)
                ),
            ]
            for output_measure, d_out in [
                (PureDP(), sp.Integer(4)),
                (RhoZCDP(), sp.Integer(4)),
                (ApproxDP(), (sp.Integer(4), sp.Integer(0))),
            ]
            if not (
                noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN
                and output_measure != RhoZCDP()
            )
        ]
    )
    def test_create_sum_measurement_with_groupby(
        self,
        input_metric: Union[SymmetricDifference, HammingDistance, IfGroupedBy],
        groupby_output_metric: Union[SumOf, RootSumOfSquared],
        output_measure: Union[PureDP, RhoZCDP],
        d_out: PrivacyBudgetInput,
        noise_mechanism: NoiseMechanism,
    ):
        """Tests that create_sum_measurement works correctly with groupby."""
        if self.groupby_columns == [] and isinstance(input_metric, IfGroupedBy):
            return
        sum_measurement = create_sum_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            output_measure=output_measure,
            measure_column="B",
            upper=sp.Integer(10),
            lower=sp.Integer(0),
            noise_mechanism=noise_mechanism,
            d_in=sp.Integer(1),
            d_out=d_out,
            groupby_transformation=GroupBy(
                input_domain=self.input_domain,
                input_metric=input_metric,
                use_l2=isinstance(groupby_output_metric, RootSumOfSquared),
                group_keys=self.group_keys,
            ),
            sum_column="sumB",
        )
        self.assertEqual(sum_measurement.input_domain, self.input_domain)
        self.assertEqual(sum_measurement.output_measure, output_measure)
        self.assertEqual(sum_measurement.privacy_function(sp.Integer(1)), d_out)
        answer = sum_measurement(self.sdf)
        self.assertIsInstance(answer, DataFrame)
        self.assertEqual(answer.columns, self.groupby_columns + ["sumB"])

    @parameterized.expand(
        [
            (
                input_metric,
                groupby_output_metric,
                output_measure,
                d_out,
                noise_mechanism,
            )
            for noise_mechanism, groupby_output_metric in [
                (NoiseMechanism.LAPLACE, SumOf(SymmetricDifference())),
                (NoiseMechanism.GEOMETRIC, SumOf(SymmetricDifference())),
                (
                    NoiseMechanism.DISCRETE_GAUSSIAN,
                    RootSumOfSquared(SymmetricDifference()),
                ),
            ]
            for input_metric in [
                SymmetricDifference(),
                HammingDistance(),
                IfGroupedBy(
                    "A", cast(Union[SumOf, RootSumOfSquared], groupby_output_metric)
                ),
            ]
            for output_measure, d_out in [
                (PureDP(), sp.Integer(4)),
                (RhoZCDP(), sp.Integer(4)),
                (ApproxDP(), (sp.Integer(4), sp.Integer(0))),
            ]
            if not (
                noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN
                and output_measure != RhoZCDP()
            )
        ]
    )
    def test_create_average_measurement_with_groupby(
        self,
        input_metric: Union[SymmetricDifference, HammingDistance, IfGroupedBy],
        groupby_output_metric: Union[SumOf, RootSumOfSquared],
        output_measure: Union[PureDP, RhoZCDP, ApproxDP],
        d_out: PrivacyBudgetInput,
        noise_mechanism: NoiseMechanism,
    ):
        """Tests that create_average_measurement works correctly with groupby."""
        if self.groupby_columns == [] and isinstance(input_metric, IfGroupedBy):
            return
        average_measurement = create_average_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            output_measure=output_measure,
            measure_column="B",
            upper=sp.Integer(10),
            lower=sp.Integer(0),
            noise_mechanism=noise_mechanism,
            d_in=sp.Integer(1),
            d_out=d_out,
            groupby_transformation=GroupBy(
                input_domain=self.input_domain,
                input_metric=input_metric,
                use_l2=isinstance(groupby_output_metric, RootSumOfSquared),
                group_keys=self.group_keys,
            ),
            average_column="AVG(B)",
        )
        self.assertEqual(average_measurement.input_domain, self.input_domain)
        self.assertEqual(average_measurement.output_measure, output_measure)
        self.assertEqual(average_measurement.privacy_function(sp.Integer(1)), d_out)
        answer = average_measurement(self.sdf)
        self.assertIsInstance(answer, DataFrame)
        self.assertEqual(answer.columns, self.groupby_columns + ["AVG(B)"])

    @parameterized.expand(
        [
            (
                input_metric,
                groupby_output_metric,
                output_measure,
                noise_mechanism,
                d_out,
                output_column,
            )
            for noise_mechanism, groupby_output_metric in [
                (NoiseMechanism.LAPLACE, SumOf(SymmetricDifference())),
                (NoiseMechanism.GEOMETRIC, SumOf(SymmetricDifference())),
                (
                    NoiseMechanism.DISCRETE_GAUSSIAN,
                    RootSumOfSquared(SymmetricDifference()),
                ),
            ]
            for input_metric in [
                SymmetricDifference(),
                HammingDistance(),
                IfGroupedBy(
                    "A", cast(Union[SumOf, RootSumOfSquared], groupby_output_metric)
                ),
            ]
            for output_measure, d_out in [
                (PureDP(), sp.Integer(4)),
                (RhoZCDP(), sp.Integer(4)),
                (ApproxDP(), (sp.Integer(4), sp.Integer(0))),
            ]
            for output_column in ["XYZ", None]
            if not (
                noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN
                and output_measure != RhoZCDP()
            )
        ]
    )
    def test_create_standard_deviation_measurement_with_groupby(
        self,
        input_metric: Union[SymmetricDifference, HammingDistance, IfGroupedBy],
        groupby_output_metric: Union[SumOf, RootSumOfSquared],
        output_measure: Union[PureDP, RhoZCDP],
        noise_mechanism: NoiseMechanism,
        d_out: PrivacyBudgetInput,
        output_column: Optional[str] = None,
    ):
        """Tests that create_standard_deviation_measurement works correctly."""
        if self.groupby_columns == [] and isinstance(input_metric, IfGroupedBy):
            return
        standard_deviation_measurement = create_standard_deviation_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            output_measure=output_measure,
            measure_column="B",
            upper=sp.Integer(10),
            lower=sp.Integer(0),
            noise_mechanism=noise_mechanism,
            d_in=sp.Integer(1),
            d_out=d_out,
            groupby_transformation=GroupBy(
                input_domain=self.input_domain,
                input_metric=input_metric,
                use_l2=isinstance(groupby_output_metric, RootSumOfSquared),
                group_keys=self.group_keys,
            ),
            keep_intermediates=False,
            standard_deviation_column=output_column,
        )
        self.assertEqual(standard_deviation_measurement.input_domain, self.input_domain)
        self.assertEqual(standard_deviation_measurement.output_measure, output_measure)
        self.assertEqual(
            standard_deviation_measurement.privacy_function(sp.Integer(1)), d_out
        )
        answer = standard_deviation_measurement(self.sdf)
        self.assertIsInstance(answer, DataFrame)
        if not output_column:
            output_column = "stddev(B)"
        self.assertEqual(answer.columns, self.groupby_columns + [output_column])
        answer.first()

    @parameterized.expand(
        [
            (
                input_metric,
                groupby_output_metric,
                output_measure,
                noise_mechanism,
                d_out,
                output_column,
            )
            for noise_mechanism, groupby_output_metric in [
                (NoiseMechanism.LAPLACE, SumOf(SymmetricDifference())),
                (NoiseMechanism.GEOMETRIC, SumOf(SymmetricDifference())),
                (
                    NoiseMechanism.DISCRETE_GAUSSIAN,
                    RootSumOfSquared(SymmetricDifference()),
                ),
            ]
            for input_metric in [
                SymmetricDifference(),
                HammingDistance(),
                IfGroupedBy(
                    "A", cast(Union[SumOf, RootSumOfSquared], groupby_output_metric)
                ),
            ]
            for output_measure, d_out in [
                (PureDP(), sp.Integer(4)),
                (RhoZCDP(), sp.Integer(4)),
                (ApproxDP(), (sp.Integer(4), sp.Integer(0))),
            ]
            for output_column in ["XYZ", None]
            if not (
                noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN
                and output_measure != RhoZCDP()
            )
        ]
    )
    def test_create_variance_measurement_with_groupby(
        self,
        input_metric: Union[SymmetricDifference, HammingDistance, IfGroupedBy],
        groupby_output_metric: Union[SumOf, RootSumOfSquared],
        output_measure: Union[PureDP, RhoZCDP],
        noise_mechanism: NoiseMechanism,
        d_out: PrivacyBudgetInput,
        output_column: Optional[str] = None,
    ):
        """Tests that create_variance_measurement works correctly with groupby."""
        if self.groupby_columns == [] and isinstance(input_metric, IfGroupedBy):
            return
        variance_measurement = create_variance_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            output_measure=output_measure,
            measure_column="B",
            upper=sp.Integer(10),
            lower=sp.Integer(0),
            noise_mechanism=noise_mechanism,
            d_in=sp.Integer(1),
            d_out=d_out,
            groupby_transformation=GroupBy(
                input_domain=self.input_domain,
                input_metric=input_metric,
                use_l2=isinstance(groupby_output_metric, RootSumOfSquared),
                group_keys=self.group_keys,
            ),
            keep_intermediates=False,
            variance_column=output_column,
        )
        self.assertEqual(variance_measurement.input_domain, self.input_domain)
        self.assertEqual(variance_measurement.output_measure, output_measure)
        self.assertEqual(variance_measurement.privacy_function(sp.Integer(1)), d_out)
        answer = variance_measurement(self.sdf)
        self.assertIsInstance(answer, DataFrame)
        if not output_column:
            output_column = "var(B)"
        self.assertEqual(answer.columns, self.groupby_columns + [output_column])
        answer.first()

    @parameterized.expand(
        [
            (input_metric, groupby_output_metric, d_out, output_measure)
            for output_measure, d_out, groupby_output_metric in [
                (PureDP(), sp.Integer(4), SumOf(SymmetricDifference())),
                (RhoZCDP(), sp.Integer(4), RootSumOfSquared(SymmetricDifference())),
                (
                    ApproxDP(),
                    (sp.Integer(4), sp.Integer(0)),
                    SumOf(SymmetricDifference()),
                ),
            ]
            for input_metric in [
                SymmetricDifference(),
                HammingDistance(),
                IfGroupedBy(
                    "A", cast(Union[SumOf, RootSumOfSquared], groupby_output_metric)
                ),
            ]
        ]
    )
    def test_create_quantile_measurement_with_groupby(
        self,
        input_metric: Union[HammingDistance, SymmetricDifference],
        groupby_output_metric: Union[SumOf, RootSumOfSquared],
        d_out: PrivacyBudgetInput,
        output_measure: Union[PureDP, RhoZCDP],
    ):
        """Tests that create_quantile_measurement works correctly with groupby."""
        if self.groupby_columns == [] and isinstance(input_metric, IfGroupedBy):
            return
        quantile_measurement = create_quantile_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            output_measure=output_measure,
            measure_column="B",
            quantile=0.5,
            upper=10,
            lower=0,
            d_in=sp.Integer(1),
            d_out=d_out,
            groupby_transformation=GroupBy(
                input_domain=self.input_domain,
                input_metric=input_metric,
                use_l2=isinstance(groupby_output_metric, RootSumOfSquared),
                group_keys=self.group_keys,
            ),
            quantile_column="MEDIAN(B)",
        )
        self.assertEqual(quantile_measurement.input_domain, self.input_domain)
        self.assertEqual(quantile_measurement.input_metric, input_metric)
        self.assertEqual(quantile_measurement.output_measure, output_measure)
        self.assertEqual(quantile_measurement.privacy_function(sp.Integer(1)), d_out)
        answer = quantile_measurement(self.sdf)
        self.assertIsInstance(answer, DataFrame)
        self.assertEqual(answer.columns, self.groupby_columns + ["MEDIAN(B)"])
        df = answer.toPandas()
        self.assertTrue(((df["MEDIAN(B)"] <= 10) & (df["MEDIAN(B)"] >= 0)).all())


class TestAggregationMeasurement(PySparkTest):
    """Tests for :mod:`tmlt.core.measurements.aggregations`."""

    def setUp(self):
        """Test setup."""
        self.input_domain = SparkDataFrameDomain(
            {"A": SparkStringColumnDescriptor(), "B": SparkIntegerColumnDescriptor()}
        )
        self.sdf = self.spark.createDataFrame([("x1", 2), ("x2", 4)], schema=["A", "B"])

    @parameterized.expand(
        [
            (input_metric, output_measure, d_out, noise_mechanism)
            for input_metric in [SymmetricDifference(), HammingDistance()]
            for noise_mechanism in [
                NoiseMechanism.LAPLACE,
                NoiseMechanism.GEOMETRIC,
                NoiseMechanism.DISCRETE_GAUSSIAN,
                NoiseMechanism.GAUSSIAN,
            ]
            for output_measure, d_out in [
                (PureDP(), sp.Integer(2)),
                (RhoZCDP(), sp.Integer(2)),
                (ApproxDP(), (sp.Integer(2), sp.Integer(0))),
            ]
            if not (
                noise_mechanism
                in [NoiseMechanism.DISCRETE_GAUSSIAN, NoiseMechanism.GAUSSIAN]
                and output_measure != RhoZCDP()
            )
        ]
    )
    def test_create_count_measurement_without_groupby(
        self,
        input_metric: Union[SymmetricDifference, HammingDistance],
        output_measure: Union[PureDP, RhoZCDP],
        d_out: PrivacyBudgetInput,
        noise_mechanism: NoiseMechanism,
    ):
        """Tests that create_count_measurement works correctly without groupby."""
        count_measurement = create_count_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            noise_mechanism=noise_mechanism,
            d_in=sp.Integer(1),
            d_out=d_out,
            output_measure=output_measure,
        )
        self.assertEqual(count_measurement.input_domain, self.input_domain)
        self.assertEqual(count_measurement.input_metric, input_metric)
        self.assertEqual(count_measurement.output_measure, output_measure)
        self.assertEqual(count_measurement.privacy_function(1), d_out)
        answer = count_measurement(self.sdf)
        self.assertIsInstance(answer, (float, int))

    @parameterized.expand(
        [
            (input_metric, output_measure, d_out, noise_mechanism)
            for input_metric in [SymmetricDifference(), HammingDistance()]
            for noise_mechanism in [
                NoiseMechanism.LAPLACE,
                NoiseMechanism.GEOMETRIC,
                NoiseMechanism.DISCRETE_GAUSSIAN,
                NoiseMechanism.GAUSSIAN,
            ]
            for output_measure, d_out in [
                (PureDP(), sp.Integer(2)),
                (RhoZCDP(), sp.Integer(2)),
                (ApproxDP(), (sp.Integer(2), sp.Integer(0))),
            ]
            if not (
                noise_mechanism
                in [NoiseMechanism.DISCRETE_GAUSSIAN, NoiseMechanism.GAUSSIAN]
                and output_measure != RhoZCDP()
            )
        ]
    )
    def test_create_count_distinct_measurement_without_groupby(
        self,
        input_metric: Union[SymmetricDifference, HammingDistance],
        output_measure: Union[PureDP, RhoZCDP],
        d_out: PrivacyBudgetInput,
        noise_mechanism: NoiseMechanism,
    ):
        """Tests create_count_distinct_measurement without groupby."""
        count_distinct_measurement = create_count_distinct_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            noise_mechanism=noise_mechanism,
            d_in=sp.Integer(1),
            d_out=d_out,
            output_measure=output_measure,
        )

        self.assertEqual(count_distinct_measurement.input_domain, self.input_domain)
        self.assertEqual(count_distinct_measurement.input_metric, input_metric)
        self.assertEqual(count_distinct_measurement.output_measure, output_measure)
        self.assertEqual(count_distinct_measurement.privacy_function(1), d_out)
        answer = count_distinct_measurement(self.sdf)
        self.assertIsInstance(answer, (int, float))

    @parameterized.expand(
        [
            (input_metric, output_measure, d_out, noise_mechanism)
            for input_metric in [SymmetricDifference(), HammingDistance()]
            for noise_mechanism in [
                NoiseMechanism.LAPLACE,
                NoiseMechanism.GEOMETRIC,
                NoiseMechanism.DISCRETE_GAUSSIAN,
                NoiseMechanism.GAUSSIAN,
            ]
            for output_measure, d_out in [
                (PureDP(), sp.Integer(4)),
                (RhoZCDP(), sp.Integer(4)),
                (ApproxDP(), (sp.Integer(4), sp.Integer(0))),
            ]
            if not (
                noise_mechanism
                in [NoiseMechanism.DISCRETE_GAUSSIAN, NoiseMechanism.GAUSSIAN]
                and output_measure != RhoZCDP()
            )
        ]
    )
    def test_create_sum_measurement_without_groupby(
        self,
        input_metric: Union[SymmetricDifference, HammingDistance],
        output_measure: Union[PureDP, RhoZCDP],
        d_out: PrivacyBudgetInput,
        noise_mechanism: NoiseMechanism,
    ):
        """Tests that create_sum_measurement works correctly without groupby."""
        sum_measurement = create_sum_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            measure_column="B",
            upper=sp.Integer(10),
            lower=sp.Integer(0),
            noise_mechanism=noise_mechanism,
            d_in=sp.Integer(1),
            d_out=d_out,
            output_measure=output_measure,
        )

        self.assertEqual(sum_measurement.input_domain, self.input_domain)
        self.assertEqual(sum_measurement.input_metric, input_metric)
        self.assertEqual(sum_measurement.output_measure, output_measure)
        self.assertEqual(sum_measurement.privacy_function(1), d_out)
        answer = sum_measurement(self.sdf)
        self.assertIsInstance(answer, (float, int))

    @parameterized.expand(
        [
            (input_metric, output_measure, d_out, noise_mechanism)
            for input_metric in [SymmetricDifference(), HammingDistance()]
            for noise_mechanism in [
                NoiseMechanism.LAPLACE,
                NoiseMechanism.GEOMETRIC,
                NoiseMechanism.DISCRETE_GAUSSIAN,
                NoiseMechanism.GAUSSIAN,
            ]
            for output_measure, d_out in [
                (PureDP(), sp.Integer(4)),
                (RhoZCDP(), sp.Integer(4)),
                (ApproxDP(), (sp.Integer(4), sp.Integer(0))),
            ]
            if not (
                noise_mechanism
                in [NoiseMechanism.DISCRETE_GAUSSIAN, NoiseMechanism.GAUSSIAN]
                and output_measure != RhoZCDP()
            )
        ]
    )
    def test_create_average_measurement_without_groupby(
        self,
        input_metric: Union[SymmetricDifference, HammingDistance],
        output_measure: Union[PureDP, RhoZCDP],
        d_out: PrivacyBudgetInput,
        noise_mechanism: NoiseMechanism,
    ):
        """Tests that create_average_measurement works correctly without groupby."""
        average_measurement = create_average_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            measure_column="B",
            upper=sp.Integer(10),
            lower=sp.Integer(0),
            noise_mechanism=noise_mechanism,
            d_in=sp.Integer(1),
            d_out=d_out,
            keep_intermediates=False,
            output_measure=output_measure,
        )

        self.assertEqual(average_measurement.input_domain, self.input_domain)
        self.assertEqual(average_measurement.input_metric, input_metric)
        self.assertEqual(average_measurement.output_measure, output_measure)
        self.assertEqual(average_measurement.privacy_function(1), d_out)
        answer = average_measurement(self.sdf)
        self.assertIsInstance(answer, (float, int))

    @parameterized.expand(
        [
            (input_metric, output_measure, d_out, noise_mechanism)
            for input_metric in [SymmetricDifference(), HammingDistance()]
            for noise_mechanism in [
                NoiseMechanism.LAPLACE,
                NoiseMechanism.GEOMETRIC,
                NoiseMechanism.DISCRETE_GAUSSIAN,
                NoiseMechanism.GAUSSIAN,
            ]
            for output_measure, d_out in [
                (PureDP(), sp.Integer(4)),
                (RhoZCDP(), sp.Integer(4)),
                (ApproxDP(), (sp.Integer(4), sp.Integer(0))),
            ]
            if not (
                noise_mechanism
                in [NoiseMechanism.DISCRETE_GAUSSIAN, NoiseMechanism.GAUSSIAN]
                and output_measure != RhoZCDP()
            )
        ]
    )
    def test_create_standard_deviation_measurement_without_groupby(
        self,
        input_metric: Union[SymmetricDifference, HammingDistance],
        output_measure: Union[PureDP, RhoZCDP],
        d_out: PrivacyBudgetInput,
        noise_mechanism: NoiseMechanism,
    ):
        """Tests that create_standard_deviation_measurement works correctly."""
        standard_deviation_measurement = create_standard_deviation_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            measure_column="B",
            upper=sp.Integer(10),
            lower=sp.Integer(0),
            noise_mechanism=noise_mechanism,
            d_in=sp.Integer(1),
            d_out=d_out,
            keep_intermediates=False,
            output_measure=output_measure,
        )

        self.assertEqual(standard_deviation_measurement.input_domain, self.input_domain)
        self.assertEqual(standard_deviation_measurement.input_metric, input_metric)
        self.assertEqual(standard_deviation_measurement.output_measure, output_measure)
        self.assertEqual(standard_deviation_measurement.privacy_function(1), d_out)
        answer = standard_deviation_measurement(self.sdf)
        self.assertIsInstance(answer, float)

    @parameterized.expand(
        [
            (input_metric, output_measure, d_out, noise_mechanism)
            for input_metric in [SymmetricDifference(), HammingDistance()]
            for noise_mechanism in [
                NoiseMechanism.LAPLACE,
                NoiseMechanism.GEOMETRIC,
                NoiseMechanism.DISCRETE_GAUSSIAN,
                NoiseMechanism.GAUSSIAN,
            ]
            for output_measure, d_out in [
                (PureDP(), sp.Integer(4)),
                (RhoZCDP(), sp.Integer(4)),
                (ApproxDP(), (sp.Integer(4), sp.Integer(0))),
            ]
            if not (
                noise_mechanism
                in [NoiseMechanism.DISCRETE_GAUSSIAN, NoiseMechanism.GAUSSIAN]
                and output_measure != RhoZCDP()
            )
        ]
    )
    def test_create_variance_measurement_without_groupby(
        self,
        input_metric: Union[SymmetricDifference, HammingDistance],
        output_measure: Union[PureDP, RhoZCDP],
        d_out: PrivacyBudgetInput,
        noise_mechanism: NoiseMechanism,
    ):
        """Tests that create_variance_measurement works correctly without groupby."""
        variance_measurement = create_variance_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            measure_column="B",
            upper=sp.Integer(10),
            lower=sp.Integer(0),
            noise_mechanism=noise_mechanism,
            d_in=sp.Integer(1),
            d_out=d_out,
            keep_intermediates=False,
            output_measure=output_measure,
        )

        self.assertEqual(variance_measurement.input_domain, self.input_domain)
        self.assertEqual(variance_measurement.input_metric, input_metric)
        self.assertEqual(variance_measurement.output_measure, output_measure)
        self.assertEqual(variance_measurement.privacy_function(1), d_out)
        answer = variance_measurement(self.sdf)
        self.assertIsInstance(answer, (int, float))

    @parameterized.expand(
        [
            (input_metric, output_measure, d_out)
            for input_metric in [SymmetricDifference(), HammingDistance()]
            for output_measure, d_out in [
                (PureDP(), sp.Integer(4)),
                (RhoZCDP(), sp.Integer(4)),
                (ApproxDP(), (sp.Integer(4), sp.Integer(0))),
            ]
        ]
    )
    def test_create_quantile_measurement_without_groupby(
        self,
        input_metric: Union[HammingDistance, SymmetricDifference],
        output_measure: Union[PureDP, RhoZCDP, ApproxDP],
        d_out: PrivacyBudgetInput,
    ):
        """Tests that create_quantile_measurement works correctly without groupby."""
        quantile_measurement = create_quantile_measurement(
            input_domain=self.input_domain,
            input_metric=input_metric,
            output_measure=output_measure,
            measure_column="B",
            quantile=0.5,
            upper=10,
            lower=0,
            d_in=sp.Integer(1),
            d_out=d_out,
            groupby_transformation=None,
            quantile_column="MEDIAN(B)",
        )
        self.assertEqual(quantile_measurement.input_domain, self.input_domain)
        self.assertEqual(quantile_measurement.input_metric, input_metric)
        self.assertEqual(quantile_measurement.output_measure, output_measure)
        self.assertEqual(quantile_measurement.privacy_function(1), d_out)
        answer = quantile_measurement(self.sdf)
        self.assertIsInstance(answer, float)
        self.assertLessEqual(answer, 10)
        self.assertGreaterEqual(answer, 0)

    @parameterized.expand(
        [
            (float("inf"), 0, 1, 0, 9223372036854775807, None),
            (
                ExactNumber(1) / ExactNumber(3),
                1 - double_sided_geometric_cmf_exact(7 - 2, 3),
                1,
                3,
                7,
                None,
            ),
            (
                ExactNumber(1) / ExactNumber(17),
                1 - double_sided_geometric_cmf_exact(10 - 2, 17),
                1,
                17,
                10,
                None,
            ),
            (
                ExactNumber(2) / ExactNumber(13),
                2
                * ExactNumber(sp.E) ** (ExactNumber(2) / ExactNumber(13))
                * (1 - double_sided_geometric_cmf_exact(50 - 2, 13)),
                2,
                13,
                50,
                "my_count_column",
            ),
        ]
    )
    def test_create_partition_selection_measurement(
        self,
        epsilon: ExactNumberInput,
        delta: ExactNumberInput,
        d_in: ExactNumberInput,
        expected_alpha: ExactNumberInput,
        expected_threshold: ExactNumberInput,
        count_column: Optional[str] = None,
    ) -> None:
        """Test create_partition_selection_measurement works correctly."""
        measurement = create_partition_selection_measurement(
            input_domain=self.input_domain,
            epsilon=epsilon,
            delta=delta,
            d_in=d_in,
            count_column=count_column,
        )
        self.assertEqual(measurement.alpha, expected_alpha)
        self.assertEqual(measurement.threshold, expected_threshold)
        self.assertEqual(measurement.input_domain, self.input_domain)
        if count_column is not None:
            self.assertEqual(measurement.count_column, count_column)
        # Check that measurement.privacy_function(d_in) = (epsilon, delta)
        measurement_epsilon, measurement_delta = measurement.privacy_function(d_in)
        self.assertEqual(measurement_epsilon, epsilon)
        self.assertEqual(measurement_delta, delta)

    @parameterized.expand(
        [
            (PureDP(), 1, "B", 0.7, 1),
            (RhoZCDP(), 1, "B", 0.7, 1),
            (ApproxDP(), (1, 0), "B", 0.7, 1),
        ]
    )
    def test_create_bound_selection_measurement(
        self,
        output_measure: Union[PureDP, RhoZCDP, ApproxDP],
        d_out: PrivacyBudgetInput,
        bound_column: str,
        threshold: float,
        d_in: ExactNumberInput = 1,
    ):
        """Test create_bound_selection_measurement works correctly."""
        measurement = create_bound_selection_measurement(
            input_domain=self.input_domain,
            output_measure=output_measure,
            d_out=d_out,
            bound_column=bound_column,
            threshold=threshold,
            d_in=d_in,
        )
        d_out = PrivacyBudget.cast(output_measure, d_out).value
        if isinstance(measurement, PureDPToRhoZCDP):
            measurement = measurement.pure_dp_measurement
            epsilon = sp.sqrt(ExactNumber(sp.Integer(2) * d_out).expr)
        elif isinstance(measurement, PureDPToApproxDP):
            measurement = measurement.pure_dp_measurement
            assert isinstance(d_out, tuple)
            epsilon = d_out[0]
        else:
            epsilon = d_out
        # Appease mypy
        if not isinstance(measurement, BoundSelection):
            raise TypeError(
                f"Expected measurement to be a BoundSelection, got {measurement}"
            )
        d_in = ExactNumber(d_in)
        self.assertEqual(measurement.input_domain, self.input_domain)
        self.assertEqual(measurement.output_measure, PureDP())
        self.assertEqual(measurement.privacy_function(d_in), epsilon)
        if d_out == float("inf"):
            expected_alpha = ExactNumber(0)
        else:
            expected_alpha = (4 / epsilon) * d_in
        self.assertEqual(measurement.alpha, expected_alpha)
        self.assertEqual(measurement.bound_column, bound_column)
        self.assertEqual(measurement.threshold, threshold)


INPUT_DOMAIN = SparkDataFrameDomain(
    {"A": SparkStringColumnDescriptor(), "B": SparkIntegerColumnDescriptor()}
)


class TestBadDelta(unittest.TestCase):
    """Tests for :mod:`tmlt.core.measurements.aggregations`."""

    @parameterized.expand(
        [
            (noise_mechanism, d_out, f)
            for noise_mechanism, d_out in [
                (NoiseMechanism.LAPLACE, (sp.Integer(1), sp.Rational(1, 2))),
                (NoiseMechanism.GEOMETRIC, (sp.Integer(1), sp.Rational(1, 2))),
                (NoiseMechanism.GAUSSIAN, (sp.Integer(1), sp.Rational(1, 2))),
                (NoiseMechanism.DISCRETE_GAUSSIAN, (sp.Integer(1), sp.Rational(1, 2))),
                (NoiseMechanism.GAUSSIAN, (sp.Integer(1), sp.Integer(0))),
                (NoiseMechanism.DISCRETE_GAUSSIAN, (sp.Integer(1), sp.Integer(0))),
            ]
            for f in [
                functools.partial(
                    create_count_measurement,
                    input_domain=INPUT_DOMAIN,
                    input_metric=SymmetricDifference(),
                    output_measure=ApproxDP(),
                ),
                functools.partial(
                    create_count_distinct_measurement,
                    input_domain=INPUT_DOMAIN,
                    input_metric=SymmetricDifference(),
                    output_measure=ApproxDP(),
                ),
                functools.partial(
                    create_sum_measurement,
                    input_domain=INPUT_DOMAIN,
                    input_metric=SymmetricDifference(),
                    measure_column="B",
                    upper=sp.Integer(10),
                    lower=sp.Integer(0),
                    output_measure=ApproxDP(),
                ),
                functools.partial(
                    create_average_measurement,
                    input_domain=INPUT_DOMAIN,
                    input_metric=SymmetricDifference(),
                    measure_column="B",
                    upper=sp.Integer(10),
                    lower=sp.Integer(0),
                    output_measure=ApproxDP(),
                ),
                functools.partial(
                    create_standard_deviation_measurement,
                    input_domain=INPUT_DOMAIN,
                    input_metric=SymmetricDifference(),
                    measure_column="B",
                    upper=sp.Integer(10),
                    lower=sp.Integer(0),
                    output_measure=ApproxDP(),
                ),
                functools.partial(
                    create_variance_measurement,
                    input_domain=INPUT_DOMAIN,
                    input_metric=SymmetricDifference(),
                    measure_column="B",
                    upper=sp.Integer(10),
                    lower=sp.Integer(0),
                    output_measure=ApproxDP(),
                ),
            ]
        ]
    )
    def test_functions_with_noise_mechanism(
        self, noise_mechanism: NoiseMechanism, d_out: PrivacyBudgetInput, f: Callable
    ) -> None:
        """Test error is raised for invalid delta/noise mechanism combination."""
        with self.assertRaises(ValueError):
            f(noise_mechanism=noise_mechanism, d_out=d_out)

    @parameterized.expand(
        [
            (d_out, f)
            for d_out in [
                (sp.Integer(1), sp.Rational(1, 2)),
                (sp.Integer(1), sp.Rational(1, 3)),
            ]
            for f in [
                functools.partial(
                    create_bound_selection_measurement,
                    input_domain=INPUT_DOMAIN,
                    bound_column="B",
                    threshold=0.5,
                    output_measure=ApproxDP(),
                ),
                functools.partial(
                    create_quantile_measurement,
                    input_domain=INPUT_DOMAIN,
                    input_metric=SymmetricDifference(),
                    measure_column="B",
                    quantile=0.5,
                    upper=10,
                    lower=0,
                    output_measure=ApproxDP(),
                ),
            ]
        ]
    )
    def test_functions_without_noise_mechanism(
        self, d_out: PrivacyBudgetInput, f: Callable
    ) -> None:
        """Test error is raised for invalid deltas."""
        with self.assertRaises(ValueError):
            f(d_out=d_out)



"""Tests `create_average_measurement` noise distributions are as expected."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

# pylint: disable=no-member, no-self-use

from typing import Dict, List, Union

import pytest
from pyspark.sql import functions as sf

from tmlt.core.measurements.aggregations import (
    NoiseMechanism,
    create_average_measurement,
    get_midpoint,
)
from tmlt.core.measures import PureDP, RhoZCDP
from tmlt.core.metrics import SymmetricDifference
from tmlt.core.utils.testing import (
    ChiSquaredTestCase,
    FixedGroupDataSet,
    KSTestCase,
    PySparkTest,
    get_noise_scales,
    get_prob_functions,
    get_sampler,
    get_values_summing_to_loc,
    run_test_using_chi_squared_test,
    run_test_using_ks_test,
)

from . import NOISE_SCALE_FUDGE_FACTOR, P_THRESHOLD, SAMPLE_SIZE


def _get_average_test_cases(noise_mechanism: NoiseMechanism) -> List[Dict]:
    """Returns average test cases.

    This returns a list of test instances specifying the sampler (that produces
    a count sample and a sum sample used to compute the average), expected locations
    for count and sum, expected noise scales and corresponding cdfs (if noise mechanism
    is Laplace) or cmfs and pmfs (if noise mechanism is not Laplace).
    """
    test_cases = []
    sum_locations: Union[List[float], List[int]]
    supports_continuous = noise_mechanism in (
        NoiseMechanism.LAPLACE,
        NoiseMechanism.GAUSSIAN,
    )
    if not supports_continuous:
        sum_locations = [100, 14]  # Must be integers
    else:
        sum_locations = [99.78, 13.63]
    count_locations = [8, 5]
    privacy_budgets = ["0.8", "0.3"]
    for sum_loc, count_loc, budget in zip(
        sum_locations, count_locations, privacy_budgets
    ):
        group_values = get_values_summing_to_loc(sum_loc, n=count_loc)
        dataset = FixedGroupDataSet(
            group_vals=group_values,
            num_groups=SAMPLE_SIZE,
            float_measure_column=supports_continuous,
        )
        measurement = create_average_measurement(
            input_domain=dataset.domain,
            input_metric=SymmetricDifference(),
            output_measure=PureDP()
            if noise_mechanism
            not in (NoiseMechanism.DISCRETE_GAUSSIAN, NoiseMechanism.GAUSSIAN)
            else RhoZCDP(),
            measure_column="B",
            lower=dataset.lower,
            upper=dataset.upper,
            noise_mechanism=noise_mechanism,
            d_out=budget,
            groupby_transformation=dataset.groupby(noise_mechanism),
            keep_intermediates=True,
            count_column="count",
            sum_column="sod",
        )

        true_answers: Dict[str, Union[float, int]] = {
            "count": len(dataset.group_vals),
            "sum": sum(dataset.group_vals),
        }
        midpoint, _ = get_midpoint(
            dataset.lower,
            dataset.upper,
            integer_midpoint=not dataset.float_measure_column,
        )
        postprocessor = lambda df, count=count_loc, midpoint=midpoint: df.withColumn(
            "sum", sf.col("sod") + sf.lit(count) * sf.lit(midpoint)
        ).select("count", "sum")
        sampler = get_sampler(measurement, dataset, postprocessor)
        noise_scales = get_noise_scales(
            agg="average",
            budget=budget,
            dataset=dataset,
            noise_mechanism=noise_mechanism,
        )
        prob_functions = get_prob_functions(noise_mechanism, true_answers)
        test_cases.append(
            {
                "sampler": sampler,
                "locations": true_answers,
                "scales": noise_scales,
                **prob_functions,
            }
        )
    return test_cases


class TestAverageNoiseDistributions(PySparkTest):
    """Distribution tests for create_average_measurement."""

    @pytest.mark.slow
    def test_average_with_laplace_noise(self):
        """Average adds noise from expected Laplace distribution."""
        cases = [
            KSTestCase.from_dict(e)
            for e in _get_average_test_cases(NoiseMechanism.LAPLACE)
        ]
        for case in cases:
            run_test_using_ks_test(case, P_THRESHOLD, NOISE_SCALE_FUDGE_FACTOR)

    @pytest.mark.slow
    def test_average_with_geometric_noise(self):
        """Average adds noise from expected geometric distribution."""
        cases = [
            ChiSquaredTestCase.from_dict(e)
            for e in _get_average_test_cases(NoiseMechanism.GEOMETRIC)
        ]
        for case in cases:
            run_test_using_chi_squared_test(case, P_THRESHOLD, NOISE_SCALE_FUDGE_FACTOR)

    @pytest.mark.slow
    def test_average_with_discrete_gaussian_noise(self):
        """Average adds noise from expected discrete Gaussian distribution."""
        cases = [
            ChiSquaredTestCase.from_dict(e)
            for e in _get_average_test_cases(NoiseMechanism.DISCRETE_GAUSSIAN)
        ]
        for case in cases:
            run_test_using_chi_squared_test(case, P_THRESHOLD, NOISE_SCALE_FUDGE_FACTOR)

    @pytest.mark.slow
    def test_average_with_gaussian_noise(self):
        """Average adds noise from expected Gaussian distribution."""
        cases = [
            KSTestCase.from_dict(e)
            for e in _get_average_test_cases(NoiseMechanism.GAUSSIAN)
        ]
        for case in cases:
            run_test_using_ks_test(case, P_THRESHOLD, NOISE_SCALE_FUDGE_FACTOR)



"""Transformations constructed by chaining other transformations."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

from typing import Any, Callable, Optional

from typeguard import typechecked

from tmlt.core.exceptions import DomainMismatchError, MetricMismatchError
from tmlt.core.transformations.base import Transformation


class ChainTT(Transformation):
    """Transformation constructed by chaining two transformations."""

    @typechecked
    def __init__(
        self,
        transformation1: Transformation,
        transformation2: Transformation,
        hint: Optional[Callable[[Any, Any], Any]] = None,
    ):
        """Constructor.

        Args:
            transformation1: Transformation to apply first.
            transformation2: Transformation to apply second.
            hint: An optional function to compute the intermediate metric value
                (after the first transformation, but before the second) for
                :meth:`~.stability_relation`. It takes in the same inputs
                as :meth:`~.stability_relation`, and is only required if
                the transformation's :meth:`~.Transformation.stability_function` raises
                :class:`NotImplementedError`.
        """
        if transformation1.output_domain != transformation2.input_domain:
            raise DomainMismatchError(
                (transformation1.output_domain, transformation2.input_domain),
                "Can not chain transformations: Mismatching domains.",
            )
        if transformation1.output_metric != transformation2.input_metric:
            raise MetricMismatchError(
                (transformation1.output_metric, transformation2.input_metric),
                "Can not chain transformations: Mismatching metrics.",
            )
        super().__init__(
            input_domain=transformation1.input_domain,
            input_metric=transformation1.input_metric,
            output_domain=transformation2.output_domain,
            output_metric=transformation2.output_metric,
        )
        self._transformation1 = transformation1
        self._transformation2 = transformation2
        self._hint = hint

    @typechecked
    def stability_function(self, d_in: Any) -> Any:
        """Returns the smallest d_out satisfied by the transformation.

        Returns M.privacy_function(T.stability_function(d_in)).

        where:

        * T1 is the first transformation applied (:attr:`~.transformation1`)
        * T2 is the second transformation applied (:attr:`~.transformation2`)

        Args:
            d_in: Distance between inputs under input_metric.

        Raises:
            NotImplementedError: If T2.stability_function(T1.stability_function(d_in))
             raises :class:`NotImplementedError`.
        """
        return self.transformation2.stability_function(
            self.transformation1.stability_function(d_in)
        )

    @typechecked
    def stability_relation(self, d_in: Any, d_out: Any) -> bool:
        """Returns True only if outputs are close under close inputs.

        Let d_mid = T1.stability_function(d_in), or hint(d_in, d_out) if
        T1.stability_function raises :class:`NotImplementedError`.

        This returns True only if the following hold:

        (1) T1.stability_relation(d_in, d_mid)
        (2) T2.stability_relation(d_mid, d_out)

        where:

        * T1 is the first transformation applied (:attr:`~.transformation1`)
        * T2 is the second transformation applied (:attr:`~.transformation2`)
        * hint is the hint passed to the constructor.

        Args:
            d_in: Distance between inputs under input_metric.
            d_out: Distance between outputs under output_metric.
        """
        self.input_metric.validate(d_in)
        self.output_metric.validate(d_out)
        try:
            d_mid = self.transformation1.stability_function(d_in)
        except NotImplementedError as e:
            if self._hint is None:
                raise ValueError(
                    "A hint is needed to check this privacy relation, because the "
                    "stability_relation of self.transformation1 raised a "
                    f"NotImplementedError: {e}"
                ) from e
            d_mid = self._hint(d_in, d_out)
        return self.transformation1.stability_relation(
            d_in, d_mid
        ) and self.transformation2.stability_relation(d_mid, d_out)

    @property
    def transformation1(self) -> Transformation:
        """Returns the first transformation being applied."""
        return self._transformation1

    @property
    def transformation2(self) -> Transformation:
        """Returns the second transformation being applied."""
        return self._transformation2

    def __call__(self, data: Any) -> Any:
        """Performs transformation1 followed by transformation2."""
        return self._transformation2(self._transformation1(data))
